/*-----------------------------------------------------------------------------
|  Copyright (c) 2012-2013, Tommy Sprinkle (tommy@tommysprinkle.com)  
|
|  All rights reserved. Redistribution and use in source and binary forms, 
|  with or without modification, are permitted provided that the following 
|  conditions are met:
|
|    * Redistributions of source code must retain the above copyright notice, 
|      this list of conditions and the following disclaimer. 
|    * Redistributions in binary form must reproduce the above copyright  
|      notice, this list of conditions and the following disclaimer in the  
|      documentation and/or other materials provided with the distribution. 
|    * Neither the name of the author nor the names of its contributors may  
|      be used toendorse or promote products derived from this software  
|      without specific prior written permission. 
|
|   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS  
|   "AS IS" AND ANYEXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
|   LIMITED TO, THE IMPLIED WARRANTIESOF MERCHANTABILITY AND FITNESS  
|   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENTSHALL THE 
|   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
|   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
|   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
|   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
|   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
|   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
|   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
|   OF THE POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fss.h"
#include "alloc.h"
#include "excp.h"
#include "finddevt.h"


static excpfile   getDataSet(char *ddn, char *dsn, char *volser, char *device);
static int dspBlock(char *dsn, excpfile iFile);
static int fmtData( int offset, char *key,   int keyln, 
				   char * data, int dataln, int showkey);
static int validateTTR(char *cttr, int *pttr);
static int procAID(void);
static int cmdShift(char * cmd, char *buf);
static int getREC(excpfile ef, struct sEXCP *excp, int tt, int r);

int main(int argc, char ** argv)
{
	excpfile iFile;
	int      rc;
	char     ddname[9];
	char     dsn[45];
	char     device[9];
	char     volser[7];

	fssInit();                                    // Initialize Full Screen Environment

	strcpy(dsn,"");                               // Clear data set name
	strcpy(device,"");                            // Clear device type
	strcpy(volser,"");                            // Clear volser

	do                                            // Loop
	{
		iFile = getDataSet(ddname, dsn, volser, device);
		if(!iFile)                                 // If no iFile returned then END 
			break;

		rc = dspBlock(dsn, iFile);                 // Call Display Routine

		excpCLOSE(iFile);                          // Close the file
		dynFree(ddname);                           // Free DD
	} while(rc != 8);                             // RC=8 - EXIT

	fssTerm();                                    // Terminate Full Screen Environment
	return 0;
}

static excpfile getDataSet(char *ddn, char *dsn, char *volser, char *device)
{
	excpfile iFile;
	int      aid;
	int      rc;
	char    *devt;
	char    *emsg;
	char     cmd[80];
	char     wk[80];


	fssReset();                                   // Start a new screen
#include "br01.pnc"

	fssSetField("DSN", dsn);                      // Initial value for DSN
	fssSetField("VOLSER", volser);                // Initial value for VOLSER

	while(1)                                      // LOOP
	{
		fssSetCursor("DSN");                       // Default Cursor Location

		fssRefresh();                              // Display Screen & Wait for Input
		fssSetField("MSG2", "");                   // Claer Error Message

		procAID();                                 // Process AID
		strcpy(cmd, fssGetField("COMMAND"));       // copy Command
		if(cmdShift(cmd, wk))                      // Shift off token
		{
			if(!strcmp("END",wk))                   // Is it END
			{
				if(cmdShift(cmd,wk))                 // No operands allowed with END
				{
					fssSetField("MSG2", "Invalid Operand for END");
					fssSetCursor("COMMAND");
				}
				else
				{
					fssReset();                       // Clean up
					return NULL;                      //    ...and reutrn to caller
				}

			}
			if(!strcmp("EXIT",wk))                  // Is command EXIT
			{
				fssReset();                          // Ignore any operands
				return NULL;                         // Clean up & return
			}
			else
			{
				fssSetField("MSG2","Invalid Command");  // else Invalid Command
				fssSetCursor("COMMAND");
			}
		}

		if(fssIsBlank(fssGetField("DSN")))         // Is DSN Blank
		{
			fssSetField("MSG2", "Enter A Dataset Name");
		}
		else
		{
			strcpy(dsn, strupr(fssGetField("DSN")));                 // Save DSN

			if(!strcmp("VTOC", dsn) || !strncmp("VTOC ", dsn, 5))    // Is it for VTOC
			{
				if(fssIsBlank(fssGetField("VOLSER")))                // Must have VOLSER & Device
				{
					fssSetField("MSG2","VOLSER Required for VTOC");
					fssSetCursor("VOLSER");
				}
				else if((devt = finddevt(strupr(fssGetField("VOLSER")))) == NULL)
				{
					fssSetField("MSG2","Invalid Device Type for Volume");
					fssSetCursor("VOLSER");
				}
				else                                                 // Got Device & VOLSER
				{
					strcpy(volser, strupr(fssGetField("VOLSER")));   // Copy Volser
					strcpy(device, devt);                            // Get Device Type

					emsg = dynAlloc( 0, ddn, volser, device);        // Alloc Dataset
					if(emsg)                                         // If allocation failed
					{
						fssSetField("MSG2", "Volume Allocation Failed"); // Err Msg
						fssSetField("MSG1", emsg);                   // Info from SVC 99
					}
					else                                             // Allocate success
					{
						iFile = excpOPENV(ddn);                      // Open File
						if(iFile)                                    // If file opened
							return iFile;                            //   return File Handle
						else                               
						{
							dynFree(ddn);                            // Free Dataset
							fssSetField("MSG2", "VTOC Failed To Open");
						}
					}
				}
			}
			//else if(fssIsBlank(fssGetField("VOLSER")))               // --- Not VTOC ---
			else
			{
				emsg = dynAlloc(dsn, ddn, 0, 0);                      // Allocate DSN
				if(emsg)
				{
					fssSetField("MSG2", "Dataset Allocation Failed");  // Allocation Failed
					fssSetField("MSG1", emsg);                         // Info from SVC 99
				}
				else
				{
					iFile = excpOPEN(ddn);                            // Open File
					if(iFile)                                         // If file opened
						return iFile;                                 //     ...return File Handle
					else
					{
						dynFree(ddn);                                 // Free Dataset
						fssSetField("MSG2", "Dataset Failed To Open");
					}
				}
			}
#if 0
			else                                                      // VOLSER Specified
			{
				if((devt = finddevt(strupr(fssGetField("VOLSER")))) == NULL)
				{
					fssSetField("MSG2","Invalid Device Type for Volume");
					fssSetCursor("VOLSER");
				}
				else                                                 // Got Device & VOLSER
				{
					strcpy(volser, strupr(fssGetField("VOLSER")));   // Copy Volser
					strcpy(device, devt);                            // Get Device Type

					emsg = dynAlloc( dsn, ddn, volser, device);      // Alloc Dataset
					if(emsg)
					{
						fssSetField("MSG2", "Dataset Allocation Failed");  // Allocation Failed
						fssSetField("MSG1", emsg);                         // Info from SVC 99
					}
					else
					{
						iFile = excpOPEN(ddn);                            // Open File
						if(iFile)                                         // If file opened
							return iFile;                                 //     ...return File Handle
						else
						{
							dynFree(ddn);                                 // Free Dataset
							fssSetField("MSG2", "Dataset Failed To Open");
						}
					}
				}
			}
#endif
		}
	}

	return 0;
}

//
//-- Validate TTR is Hex
//
static int validateTTR(char *cttr, int *pttr)
{
	char *p;
	int   i;

	strupr(cttr);                                 // Translate to Uppercase

	if(strlen(cttr) != 6)                         // Must be 6 characters
		return -1;

	p = cttr;                                     // ptr to string
	i = 0;                                        // clear TTR value
	while(*p)
	{
		i = i<<4;                                  // shift over 4 bits (1 hex char)
		if(!isxdigit(*p))                          // verify hex digit
			return -1;
		if(isdigit(*p))                            // if digit (0-9)
			i += (*p - '0');                        // add to ttr
		else                                       // else A-F
			i += (*p - 'A') +10;                    // add to ttr
		p++;
	}

	*pttr = i;                                    // return converted ttr value
	return 0;
} 

//
//-- shift off a command token
//
static int cmdShift(char * cmd, char *buf)
{
	int   len;
	char *p;
	char *p1;

	p = cmd;                                      // start of command string
	while(*p == ' ')                              // skip over blanks
		p++;

	if(!*p)                                       // if end of string
		return 0;

	p1 = p;                                       // remember start of token
	while(*p && *p != ' ')                        // find end of token
		p++;

	len = p - p1;                                 // token length
	memcpy(buf, p1, len);                         // copy token
	*(buf+len) = '\0';

	strcpy(cmd, p);                               // shift over past removed token
	return 1;
}

//
//-- process AID
//
static int procAID(void)
{
	int    aid;
	char   wk[80];

	wk[0] = '\0';
	aid = fssGetAID();                            // Get AID value
	if( aid == fssPFK03 || aid == fssPFK15)       // PFK 3 or 15
		strcpy(wk, "END ");
	if( aid == fssPFK04 || aid == fssPFK16)       // PFK 4 or 16
		strcpy(wk, "EXIT ");
	else if(aid == fssPFK08 || aid == fssPFK20)   // PFK 8 or 20
		strcpy(wk, "DOWN ");
	else if(aid == fssPFK07 || aid == fssPFK19)   // PFK 7 or 19
		strcpy(wk, "UP ");
	else if(aid == fssPFK10 || aid == fssPFK22)   // PFK 10 or 22
		strcpy(wk, "PREV");
	else if(aid == fssPFK11 || aid == fssPFK23)   // PFK 11 or 23
		strcpy(wk, "NEXT ");

	strcat(wk, fssGetField("COMMAND"));           // Append contents of COMMAND field
	fssSetField("COMMAND", strupr(wk));           // Copy back to COMMAND

	return 0;
}



//
//-- Is string all numeric digits
//
static int isNumber(char *str)
{
	while(*str)
	{
		if(!isdigit(*str))
			return 0;
		str++;
	}
	return 1;
}



static int dspBlock(char *dsn, excpfile iFile)
{
	int    aid;
	int    rc;
	int    ttr;
	int    tt;
	int    r;
	int    sttr;
	int    nextTTR;
	int    offset;
	int    next;
	char  *cttr;
	char  *showkey;

	struct sEXCP excp;
	int    i1;
	int    i2;
	char   wk[80];
	char   cmd[80];


	fssReset();
#include "br02.pnc"

	fssSetField("DSN", dsn);
	fssSetField("TTR", "000001");
	fssSetField("SHOWKEY", "Y");
	fssSetField("CCHHR", "----------");
	fssSetCursor("TTR");

	offset = 0;
	sttr   = 0;
	next   = 0;

	getREC(0, 0, 0, 0);

	while(1)
	{
		cttr = fssGetField("TTR");
		if( next == 0 && validateTTR(cttr, &ttr))
		{
			fssSetField("MSG2", "Invalid TTR Value");
		} 
		else
		{
			if(next)
				ttr = nextTTR;

			if(ttr != sttr)
			{
				do
				{
					//               rc = excpREAD(iFile, &excp, (ttr>>8), (ttr & 0xFF) ); 
					//               rc = excpTREAD(iFile, &excp, (ttr>>8) );
					rc = getREC(iFile, &excp, (ttr>>8), (ttr & 0xFF) ); 
					if(!next)
						break;
					if(rc == 4)
					{
						ttr = (((ttr + 0x100) & 0xFFFF00) | 0x01);  // Next track
						sprintf(wk, "%6.6X", ttr);
						fssSetField("TTR", wk);
					}
				} while(rc ==4);

				sttr   = ttr;
				offset = 0;
				next   = 0;
			}

			if(rc == 0)   // Read OK
			{
				ttr  = (ttr & 0xFFFF00) + *((excp.mbbcchhr)+7);
				sttr = ttr;
				sprintf(wk, "%6.6X", ttr);
				fssSetField("TTR", wk);

				i1 = *((int *) (excp.mbbcchhr));
				i2 = *((int *) ((excp.mbbcchhr)+4));
				sprintf(wk, "%8.8X%8.8X",i1,i2);
				fssSetField("CCHHR",wk+6);
			}
			else if(rc == 4)
			{
				excp.keylen = 0;
				excp.datalen = 0;
				fssSetField("MSG2", "No Record Found");
			}
			else if(rc == 8)  // EOF
			{
				excp.keylen = 0;
				excp.datalen = 0;
				fssSetField("MSG2", "*** EOF ***");
			}
			else if(rc == 12)
			{
				excp.keylen = 0;
				excp.datalen = 0;
				fssSetField("MSG2", "TTR Outside Data Set Extent");
			}
			else if(rc == 16)
			{
				excp.keylen = 0;
				excp.datalen = 0;
				fssSetField("MSG2", "I/O Error Reading Dataset");
			}
			else
			{
				excp.keylen = 0;
				excp.datalen = 0;
				fssSetField("MSG2", "Error Reading Dataset");
			}

			fssSetField("KEYL", itoa(excp.keylen, wk, 10));
			fssSetField("DATAL", itoa(excp.datalen, wk, 10));
			showkey = strupr(fssGetField("SHOWKEY"));
			fmtData( offset, excp.key, excp.keylen, excp.data, excp.datalen,
				(*showkey == 'Y') );
		}


		fssRefresh();
		fssSetField("MSG2", "");

		procAID();           // Process AID - prepend PFK command

		fssSetCursor("TTR");
		strcpy(cmd, fssGetField("COMMAND"));
		if(cmdShift(cmd, wk))
		{
			if(!strcmp("END",wk))
			{
				if(cmdShift(cmd,wk))
				{
					fssSetField("MSG2", "Invalid Operand for END");
					fssSetCursor("COMMAND");
				}
				else
				{
					fssReset();
					return 4;
				}

			}
			if(!strcmp("EXIT",wk))
			{
				fssReset();
				return 8;
			}
			else if(!strcmp("DOWN",wk))
			{
				if(cmdShift(cmd,wk))
				{
					if(!isNumber(wk))
					{
						fssSetField("MSG2", "Invalid Scroll Amount");
						fssSetCursor("COMMAND");
					}
					else
					{
						offset += atoi(wk) * 16;
						if(offset >= excp.datalen)
							offset = ((excp.datalen-(18*16))-1) & 0xFFFFFFF0;
						if(offset < 0)
							offset = 0;
						fssSetField("COMMAND", "");
					}
				}
				else
				{
					offset += 19 * 16;
					if(offset >= excp.datalen)
						offset = ((excp.datalen-(18*16))-1) & 0xFFFFFFF0;
					if(offset < 0)
						offset = 0;
					fssSetField("COMMAND", "");
				}
			}
			else if(!strcmp("UP",wk))
			{
				if(cmdShift(cmd,wk))
				{
					if(!isNumber(wk))
					{
						fssSetField("MSG2", "Invalid Scroll Amount");
						fssSetCursor("COMMAND");
					}
					else
					{
						offset -= atoi(wk) * 16;
						if(offset <0)
							offset = 0;
						fssSetField("COMMAND", "");
					}
				}
				else
				{
					offset -= 19 * 16;
					if(offset < 0)
						offset = 0;
					fssSetField("COMMAND", "");
				}
			}
			else if(!strcmp("NEXT",wk))
			{
				next = 1;
				ttr += 1;
				nextTTR = ttr;
				sprintf(wk, "%6.6X", nextTTR);
				fssSetField("TTR", wk);
				fssSetField("COMMAND", "");
			}
			else if(!strcmp("PREV",wk))
			{
				tt = (ttr >> 8);
				r  = (ttr & 0xFF);
				r--;
				if(r < 1)
				{
					r = 255;
					tt--;
					if(tt < 0)
					{
						tt = 0;
						r  = 1;
					}
				}
				nextTTR = (tt << 8) + r;
				next = 1;
				sprintf(wk, "%6.6X", nextTTR);
				fssSetField("TTR", wk);
				fssSetField("COMMAND", "");
			}
			else
			{
				fssSetField("MSG2","Invalid Command");
				fssSetCursor("COMMAND");
			}
		}
	}

	return 0;
}


static int fmtData( int offset, char *key, int keyln, 
				   char * data, int dataln, int showkey)
{
	int   i;
	int   j;
	int   l;
	int   doff;
	unsigned int ui1;
	unsigned int ui2;
	unsigned int ui3;
	unsigned int ui4;
	char *p;
	char *w;
	char  wkbuf[64];
	char  wkbuf2[64];
	char  fld[16];

	if(showkey)
	{
		p = key;
		l = keyln + dataln;
	}
	else
	{
		p    = data;
		l    = dataln;
	}

	doff = offset;
	p   += offset;
	l   -= offset;

	for( i = 1; i < 20 && l; i++)
	{
		if(l >= 16)
		{
			sprintf(wkbuf, "%6.6X", doff);
			sprintf(fld,   "OFF%2.2d", i);
			fssSetField(fld, wkbuf);

			ui1 = *( (unsigned int *) p);
			ui2 = *( (unsigned int *) (p+4));
			ui3 = *( (unsigned int *) (p+8));
			ui4 = *( (unsigned int *) (p+12));
			sprintf(wkbuf, "%8.8X %8.8X %8.8X %8.8X",ui1, ui2, ui3, ui4);
			sprintf(fld,   "HEX%2.2d", i);
			fssSetField(fld, wkbuf);

			for(j=0; j < 16; j++)
			{
				if(isprint( *(p+j) ) )
					*(wkbuf+2+j) = *(p+j);
				else
					*(wkbuf+2+j) = '.';
			}
			*(wkbuf)    = '*';
			*(wkbuf+1)  = ' ';
			*(wkbuf+18) = ' ';
			*(wkbuf+19) = '*';
			*(wkbuf+20) = '\0';
			sprintf(fld, "PRT%2.2d", i);
			fssSetField(fld, wkbuf);

			doff += 16;
			l    -= 16;
			p    += 16;
		} // if(l > 16 )
		else
		{
			sprintf(wkbuf, "%6.6X", doff);
			sprintf(fld,   "OFF%2.2d", i);
			fssSetField(fld, wkbuf);

			j = l;
			w = wkbuf;
			ui1 = *( (unsigned int *) p);
			sprintf(wkbuf2, "%8.8X", ui1);
			if(j >3)
			{
				memcpy(w, wkbuf2, 8);
				w[8] = ' ';
				j -= 4; 
				w += 9;
				ui1 = *( (unsigned int *) (p+4));
				sprintf(wkbuf2, "%8.8X", ui1);
				if(j > 3)
				{
					memcpy(w, wkbuf2, 8);
					w[8] = ' ';
					j -= 4;
					w += 9;
					ui1 = *( (unsigned int *) (p+8));
					sprintf(wkbuf2, "%8.8X", ui1);
					if( j > 3)
					{
						memcpy(w, wkbuf2, 8);
						w[8] = ' ';
						j -= 4;
						w += 9;
						ui4 = *( (unsigned int *) (p+12));
					}
				}
			}
			j *= 2;
			memcpy(w, wkbuf2, j);
			w[j] = '\0';
			sprintf(fld,   "HEX%2.2d", i);
			fssSetField(fld, wkbuf);

			memset(wkbuf, ' ', sizeof(wkbuf));

			for(j=0; j < l; j++)
			{
				if(isprint( *(p+j) ) )
					*(wkbuf+2+j) = *(p+j);
				else
					*(wkbuf+2+j) = '.';
			}
			*(wkbuf)     = '*';
			*(wkbuf+1)   = ' ';
			*(wkbuf+18)  = ' ';
			*(wkbuf+19)  = '*';
			*(wkbuf+20)  = '\0';
			sprintf(fld, "PRT%2.2d", i);
			fssSetField(fld, wkbuf);

			break;
		}
	}

	if( l )
		i++;
	for( ; i < 20; i++)
	{
		sprintf(fld,   "OFF%2.2d", i);
		fssSetField(fld, "");
		sprintf(fld,   "HEX%2.2d", i);
		fssSetField(fld, "");
		sprintf(fld, "PRT%2.2d", i);
		fssSetField(fld, "");
	}

	return 0;
}


//
//-- Use Track Read to get Data Blocks
//
static int getREC(excpfile ef, struct sEXCP *excp, int tt, int r)
{
	static int curTrack     = -1;
	static int blocks       =  0;
	static int excpRC       =  0;
	static struct sEXCP sx;
	static char *blkptr[255];

	char *p;
	int   kl;
	int   dl;
	int   len;

	if(ef == NULL)
	{
		curTrack = -1;
		blocks   =  0;
		excpRC   =  0;
		return 0;
	}

	if(tt != curTrack)
	{
		curTrack = -1;

		excpRC = excpTREAD(ef, &sx, tt);
		if(excpRC)
			return excpRC;

		curTrack = tt;

		// Determine number of blocks on track

		len = sx.datalen;
		p   = sx.data;
		blocks = 0;

		while(len > 0)
		{
			blocks++;
			blkptr[blocks] = p;
			kl = *(p+5);
			dl = (*(p+6) << 8) + *(p+7);
			p   += 8 + kl + dl;
			len -= 8 + kl + dl;
		}
	}

	if(excpRC)
		return excpRC;

	if(r == 255)
		r = blocks;

	if(r > blocks)
		return 4;

	p = blkptr[r];
	memcpy((excp->mbbcchhr)+3, p, 5);
	excp->keylen  = *(p+5);
	excp->datalen = (*(p+6) << 8) + *(p+7);
	excp->count   = p;
	excp->key     = p + 8;
	excp->data    = p + 8 + excp->keylen;

	if( (excp->keylen + excp->datalen) == 0)
		return 8;                              // End Of File

	return 0;
}
