/*-----------------------------------------------------------------------------
|  Copyright (c) 2012-2013, Tommy Sprinkle (tommy@tommysprinkle.com)  
|
|  All rights reserved. Redistribution and use in source and binary forms, 
|  with or without modification, are permitted provided that the following 
|  conditions are met:
|
|    * Redistributions of source code must retain the above copyright notice, 
|      this list of conditions and the following disclaimer. 
|    * Redistributions in binary form must reproduce the above copyright  
|      notice, this list of conditions and the following disclaimer in the  
|      documentation and/or other materials provided with the distribution. 
|    * Neither the name of the author nor the names of its contributors may  
|      be used toendorse or promote products derived from this software  
|      without specific prior written permission. 
|
|   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS  
|   "AS IS" AND ANYEXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
|   LIMITED TO, THE IMPLIED WARRANTIESOF MERCHANTABILITY AND FITNESS  
|   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENTSHALL THE 
|   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
|   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
|   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
|   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
|   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
|   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
|   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
|   OF THE POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fss.h>
#include <chainr.h>

struct sEnqRec
{
	struct sEnqRec *next;
	struct sEnqRec *prev;

	int             key;
	char            asid[4];
	char            qname[8];
	char            rname[44];
	char            jobname[8];
	char            wait[4];
	char            flag;
};

static void enqInfo(CHAINR enqChain, char *selQname, char *selRname, char *selJobname, char* tmpDS);
static int contention(CHAINR enqChain, CHAINR contChain);
static int findENQ(CHAINR enqChain, CHAINR contChain, char *qname, char *rname);
static int isOnChain(CHAINR contChain, char *qname, char *rname);
static int displayChain(CHAINR contChain);
static int cmdShift(char * cmd, char *buf);
static int procAID(void);
static struct sEnqRec * scroll(CHAINR chnr, struct sEnqRec *ptr, int dir, int amt);
static int isNumber(char *str);


#define DASHES "---------------------------------"

int main(int argc, char ** argv)
{
	CHAINR enqChain;
	CHAINR contChain;
	struct sEnqRec *rec;

	int    exit_rc;

	char   QNAME[9];
	char   JOBNAME[9];
	char   RNAME[45];
	char   TEMPDS[2];

	char   cmd[80];
	char   wk[80];

	fssInit();

	exit_rc = 0;

	enqChain  = chainNewInt(sizeof(struct sEnqRec));
	contChain = chainNewInt(sizeof(struct sEnqRec));

	strcpy(QNAME,"");
	strcpy(RNAME,"");
	strcpy(JOBNAME,"");
	strcpy(TEMPDS,"N");


	do
	{
		fssReset();
#include "senqpnl.pnc"

		fssSetCursor("COMMAND");                        // Default Cursor Position
		fssSetField("MSG1",DASHES);                     // Clear Error Msg
		fssSetColor("MSG1",0);

		fssSetField("QNAME",QNAME);
		fssSetField("RNAME",RNAME);
		fssSetField("JOBNAME",JOBNAME);
		fssSetField("TEMPDS",TEMPDS);

		fssSetField("COMMAND", "1");

		do												// Main Terminal I/O Loop
		{
			fssRefresh();                               // Update Screen
			fssSetCursor("COMMAND");                    // Default Cursor Position
			fssSetField("MSG1",DASHES);                 // Default Error Msg
			fssSetColor("MSG1",0);                      // Default Color

			procAID();                                  // Process AID
			strcpy(cmd, fssGetField("COMMAND"));        // Copy Command
			if(cmdShift(cmd, wk))                       // Shift off first token
			{
				if(!strcmp("END",wk))                   // END
				{
					fssReset();
					exit_rc = 4;						// Set RC=4 and exit
					break;
				}
				else if(!strcmp("EXIT",wk))				// EXIT
				{
					fssReset();
					exit_rc = 8;                        // Set RC=8 and exit
					break;
				}
				else if(!strcmp("1",wk))
				{
					strcpy(QNAME,strupr(fssGetField("QNAME")));
					strcpy(RNAME,strupr(fssGetField("RNAME")));
					strcpy(JOBNAME,strupr(fssGetField("JOBNAME")));
					strcpy(TEMPDS,strupr(fssGetField("TEMPDS")));
					if( *TEMPDS != 'Y')
						strcpy(TEMPDS,"N");
					fssSetField("TEMPDS",TEMPDS);

					enqInfo(enqChain, QNAME, RNAME, JOBNAME, TEMPDS);
					displayChain(enqChain);
					fssSetField("Command","");
					break;
				}
				else if(!strcmp("2",wk))
				{
					enqInfo(enqChain, "", "", "", "Y");
					contention(enqChain, contChain);
					displayChain(contChain);
					fssSetField("Command","");
					break;
				}
				else                                    // Anything Else - Invalid Command
				{
					fssSetField("MSG1","Invalid Command");
					fssSetColor("MSG1",fssYELLOW);
					fssSetCursor("COMMAND");
				}
			}
		} while(1);
	} while(exit_rc == 0);

	chainFree(enqChain);
	chainFree(contChain);
	fssTerm();

	return 0;
}


static int displayChain(CHAINR enqChain)
{
	struct sEnqRec *rec;
	struct sEnqRec *scrollPtr;
	int             i;
	int             exit_rc;
	int             color;
	char            fld[16];
	char			cmd[80];
	char			wk[80];


	fssReset();
#include "senqpnl1.pnc"

	fssSetCursor("COMMAND");								// Default Cursor Position
	fssSetField("MSG1",DASHES);							// Clear Error Msg
	fssSetColor("MSG1",0);

	scrollPtr = (struct sEnqRec *) chainFirst(enqChain);

	do
	{
		rec = scrollPtr;
		i   = 1;

		while(rec != NULL && i < 22)
		{
			if( *(rec->qname) == '-')
				color = fssTURQ;
			else if( *(rec->wait) == 'W')
				color = fssRED;
			else if( *(rec->qname) == ' ')
				color = fssYELLOW;
			else if( rec->flag != 'S')
				color = fssPINK;
			else
				color = fssGREEN;

			sprintf(fld,"QNAME%2.2d",i);
			fssSetField(fld,rec->qname);
			fssSetColor(fld,color);

			sprintf(fld,"RNAME%2.2d",i);
			fssSetField(fld,rec->rname);
			fssSetColor(fld,color);

			sprintf(fld,"FLG%2.2d",i);
			fssSetField(fld,&rec->flag);
			fssSetColor(fld,color);

			sprintf(fld,"WT%2.2d",i);
			fssSetField(fld,rec->wait);
			fssSetColor(fld,color);

			sprintf(fld,"AS%2.2d",i);
			fssSetField(fld,rec->asid);
			fssSetColor(fld,color);

			sprintf(fld,"JOBN%2.2d",i);
			fssSetField(fld,rec->jobname);
			fssSetColor(fld,color);

			i++;
			rec = rec->next;
		}

		while(i < 22)
		{
			sprintf(fld,"QNAME%2.2d",i);
			fssSetField(fld,"");

			sprintf(fld,"RNAME%2.2d",i);
			fssSetField(fld,"");

			sprintf(fld,"FLG%2.2d",i);
			fssSetField(fld,"");

			sprintf(fld,"WT%2.2d",i);
			fssSetField(fld,"");

			sprintf(fld,"AS%2.2d",i);
			fssSetField(fld,"");

			sprintf(fld,"JOBN%2.2d",i);
			fssSetField(fld,"");

			i++;
		}

		fssRefresh();
		fssSetCursor("COMMAND");						// Default Cursor Position
		fssSetField("MSG1",DASHES);						// Clear Error Msg
		fssSetColor("MSG1",0);

		procAID();
		strcpy(cmd, fssGetField("COMMAND"));        // Copy Command
		if(cmdShift(cmd, wk))                       // Shift off first token
		{
			if(!strcmp("END",wk))                   // END 
			{
				fssReset();                     // Exit...
				exit_rc = 4;
				break;
			}
			if(!strcmp("EXIT",wk))                  // EXIT
			{
				fssReset();                         // Exit...
				exit_rc = 8;
				break;
			}
			else if(!strcmp("DOWN",wk))             // DOWN
			{
				if(cmdShift(cmd,wk))                // Check for Scroll Amt
				{
					strupr(wk);
					if(!strcmp(wk,"MAX") || !strcmp(wk,"M"))   // MAX
					{
						scrollPtr = scroll(enqChain, scrollPtr, 0, -1);  // Scroll Down Max
						fssSetField("COMMAND","");
					}
					else if(!isNumber(wk))                     // Number Specified ?
					{
						fssSetField("MSG1", "Invalid Scroll Amount");  // No - Invalid
						fssSetColor("MSG1",fssYELLOW);         // Set Error Msg
						fssSetCursor("COMMAND");
					}
					else
					{
						scrollPtr = scroll(enqChain, scrollPtr, 0, atoi(wk));  // Scroll Down
						fssSetField("COMMAND","");
					}
				}
				else
				{
					scrollPtr = scroll(enqChain, scrollPtr, 0, 0);   // Default, Scroll 1 Page
					fssSetField("COMMAND","");
				}
			}
			else if(!strcmp("UP",wk))               // UP
			{
				strupr(wk);
				if(cmdShift(cmd,wk))                // Check for Scroll Amt
				{
					if(!strcmp(wk,"MAX") || !strcmp(wk,"M"))  // Max
					{
						scrollPtr = scroll(enqChain, scrollPtr, 1, -1);  // Scroll Up Max
						fssSetField("COMMAND","");
					}
					else if(!isNumber(wk))                    // If not Number, then Error
					{
						fssSetField("MSG1", "Invalid Scroll Amount");
						fssSetColor("MSG1",fssYELLOW);
						fssSetCursor("COMMAND");
					}
					else
					{
						scrollPtr = scroll(enqChain, scrollPtr, 1, atoi(wk)); // Scroll Down
						fssSetField("COMMAND","");
					}
				}
				else
				{
					scrollPtr = scroll(enqChain, scrollPtr, 1, 0);  // Default - One Page
					fssSetField("COMMAND","");
				}
			}
			else if(!strcmp("TOP",wk))              // TOP 
			{
				scrollPtr = scroll(enqChain, scrollPtr, 1, -1); // Scroll Up Max
				fssSetField("COMMAND","");
			}
			else if(!strcmp("BOT",wk) || !strcmp("BOTTOM",wk))  // BOTTOM
			{
				scrollPtr = scroll(enqChain, scrollPtr, 0, -1); // Scroll Down Max
				fssSetField("COMMAND","");
			}	
			else                                             // Else - Invalid Command
			{
				fssSetField("MSG1","Invalid Command");
				fssSetColor("MSG1",fssYELLOW);
				fssSetCursor("COMMAND");
			}
		}

	} while(1);

	fssReset();

	return exit_rc;
}


static void enqInfo(CHAINR enqChain, char *selQname, char *selRname, char *selJobname, char *tmpDS)
{
	unsigned char *CVT;
	unsigned char *MajQCB;
	unsigned char *MinQCB;
	unsigned char *QEL;
	unsigned char *ASVT;
	unsigned char *ASCB;
	unsigned char *CSCB;
	unsigned char *nxtCSCB;

	struct sEnqRec *rec;

	int            i;
	int            asid;
	int            iflg;
	int            minlen;
	int            match;
	int            ln;
	int            key;

	char           wk[16];
	char           qname[8];
	char           rname[44];
	char           jobname[8];
	char           wait[4];
	char           flag;

	chainReset(enqChain);
	key = 0;

	CVT    = *((unsigned char **) (0x10));
	MajQCB = *((unsigned char **) (CVT+0x280));

	while(MajQCB != NULL)
	{
		MinQCB = *((unsigned char **) (MajQCB+0x08));

		while(MinQCB != NULL)
		{
			memcpy(qname, MajQCB+0x10, 8);

			minlen = *(MinQCB+0x10);
			if(minlen > 44)
				minlen = 44;
			memcpy(rname, MinQCB+0x14, minlen);
			if(minlen < 44)
				rname[minlen] = 0x00;

			QEL = *((unsigned char **) (MinQCB+0x08));

			while(QEL != NULL)
			{
				asid = (*(QEL+0x0E) << 8) + *(QEL+0x0F);

				iflg = (*(QEL+0x0C) & 0x90);
				if(iflg == 0x80)
					flag = 'S';
				else if(iflg == 0x10)
					flag = 'X';
				else if(iflg == 0x90)
					flag = 'R';
				else
					flag = 'E';

				ASVT = *((unsigned char **) (CVT+0x22C));
				ASCB = *((unsigned char **) (ASVT+0x20C+(asid*4)));
				CSCB = *((unsigned char **) (ASCB+0x38));
				if( *(CSCB+0x1C) == 0x03 )  // INIT ?
				{
					nxtCSCB = *((unsigned char **) (CSCB+0x00));
					while(1)
					{
						if(nxtCSCB == NULL)
						{
							memcpy(jobname,CSCB+0x08, 8);
							break;
						}
						i = (*(nxtCSCB+0x1E) << 8) + *(nxtCSCB+0x1F);
						if( asid == i )
						{
							memcpy(jobname,nxtCSCB+0x08, 8);
							break;
						}
						nxtCSCB = *((unsigned char **) (nxtCSCB+0x00));
					}
				}
				else
					memcpy(jobname,CSCB+0x08, 8);
				if(asid == 1)
					memcpy(jobname,"MASTER  ",8);

				if( ((*(QEL+8)) & 0x80) == 0x00 && 
					((*(QEL+0x16) > 0) || (*(QEL+0x17) > 0))   )
				{
					memcpy(wait,"Wait",4);
				}
				else
				{
					memset(wait,0x00,4);
				}

				match = 1;
				if(!fssIsBlank(selQname))
				{
					ln = strlen(selQname);
					if(memcmp(selQname, qname, ln))
						match = 0;
				}
				if(!fssIsBlank(selRname))
				{
					ln = strlen(selRname);
					if(memcmp(selRname, rname, ln))
						match = 0;
				}
				if(!fssIsBlank(selJobname))
				{
					ln = strlen(selJobname);
					if(memcmp(selJobname, jobname, ln))
						match = 0;
				}
				if(*tmpDS != 'Y')
				{
					if(!memcmp("SYSDSN  ",qname,   8) &&
					   !memcmp("SYS"     ,rname,   3) &&
					   !memcmp(".T"      ,rname+8, 2) &&
					   !memcmp(".RA000." ,rname+16,7)    )
						match = 0;
				}

				if(match == 1)
				{
					key++;
					rec = (struct sEnqRec *) chainAdd(enqChain,key);

					sprintf(wk,"%4.4X",asid);
					memcpy(rec->asid,    wk,      4);
					memcpy(rec->qname,   qname,   8);
					memcpy(rec->rname,   rname,  44);
					memcpy(rec->jobname, jobname, 8);
					memcpy(rec->wait,    wait,    4);
					rec->flag = flag;
				}

				QEL = *((unsigned char **) (QEL+0x00));
			}

			MinQCB = *((unsigned char **) (MinQCB+0x00));
		}

		MajQCB = *((unsigned char **) (MajQCB+0x00));
	}
}


static int isOnChain(CHAINR contChain, char * qname, char * rname)
{
	struct sEnqRec *rec;

	rec = (struct sEnqRec *) chainFirst(contChain);

	while(rec != NULL)
	{
		if(!memcmp(qname,rec->qname, 8) &&
			!memcmp(rname,rec->rname,44)    )
			return 1;
		rec = rec->next;
	}
	return 0;
}



static int findENQ(CHAINR enqChain, CHAINR contChain, char *qname, char *rname)
{
	struct sEnqRec *rec;
	struct sEnqRec *newRec;
	int             key;

	rec = (struct sEnqRec *) chainLast(contChain);
	if(rec != NULL)
		key = rec->key;
	else
		key = 1;

	rec = (struct sEnqRec *) chainFirst(enqChain);

	while(rec != NULL)
	{
		if(!memcmp(qname, rec->qname,  8) &&
			!memcmp(rname, rec->rname, 44)    )
		{
			key++;
			newRec = (struct sEnqRec *) chainAdd(contChain, key);
			memcpy(newRec->qname,   rec->qname,   8);
			memcpy(newRec->rname,   rec->rname,  44);
			memcpy(newRec->jobname, rec->jobname, 8);
			memcpy(newRec->wait,    rec->wait,    4);
			memcpy(newRec->asid,    rec->asid,    4);
			newRec->flag = rec->flag;
		}
		rec = rec->next;
	}

	return 0;
}


static int contention(CHAINR enqChain, CHAINR contChain)
{
	struct sEnqRec *rec;
	struct sEnqRec *wrec;
	int             rc;
	int             key;

	rc = 0;
	chainReset(contChain);
	rec = (struct sEnqRec *) chainFirst(enqChain);

	while(rec)
	{
		if(*(rec->wait) != 0x00)
		{
			if(!isOnChain(contChain, rec->qname, rec->rname))
			{
				findENQ(enqChain, contChain, rec->qname, rec->rname);
				rc = 1;
				wrec = (struct sEnqRec *) chainLast(contChain);
				if(wrec == NULL)
					key = 1;
				else
					key = wrec->key + 1;
				wrec = (struct sEnqRec *) chainAdd(contChain,key);
				memset(wrec->qname,   '-',  8);
				memset(wrec->rname,   '-', 44);
				memset(wrec->jobname, '-',  8);
				memset(wrec->asid,    '-',  4);
				memset(wrec->wait,    '-',  4);
				wrec->flag = '-';
			}
		}
		rec = rec->next;
	}

	if(chainCount(contChain) == 0)
	{
		wrec = (struct sEnqRec *) chainAdd(contChain,1);
		memset(wrec->qname,   ' ',  8);
		memset(wrec->rname,   ' ', 44);
		memcpy(wrec->rname+22,"No ENQ Contention",17);
		memset(wrec->jobname, ' ',  8);
		memset(wrec->asid,    ' ',  4);
		memset(wrec->wait,    ' ',  4);
		wrec->flag = ' ';
	}
	return rc;
}


//
//-- shift off a command token
//
static int cmdShift(char * cmd, char *buf)
{
	int   len;
	char *p;
	char *p1;

	p = cmd;                                      // start of command string
	while(*p == ' ')                              // skip over blanks
		p++;

	if(!*p)                                       // if end of string
		return 0;

	p1 = p;                                       // remember start of token
	while(*p && *p != ' ')                        // find end of token
		p++;

	len = p - p1;                                 // token length
	memcpy(buf, p1, len);                         // copy token
	*(buf+len) = '\0';

	strcpy(cmd, p);                               // shift over past removed token
	return 1;
}

//
//-- process AID
//
static int procAID(void)
{
	int    aid;
	char   wk[80];

	wk[0] = '\0';
	aid = fssGetAID();                            // Get AID value
	if( aid == fssPFK03 || aid == fssPFK15)       // PFK 3 or 15
		strcpy(wk, "END ");
	if( aid == fssPFK04 || aid == fssPFK16)       // PFK 4 or 16
		strcpy(wk, "EXIT ");
	else if(aid == fssPFK08 || aid == fssPFK20)   // PFK 8 or 20
		strcpy(wk, "DOWN ");
	else if(aid == fssPFK07 || aid == fssPFK19)   // PFK 7 or 19
		strcpy(wk, "UP ");

	strcat(wk, fssGetField("COMMAND"));           // Append contents of COMMAND field
	fssSetField("COMMAND", strupr(wk));           // Copy back to COMMAND

	return 0;
}

static struct sEnqRec * scroll(CHAINR chnr, struct sEnqRec *ptr, int dir, int amt)  
{                           //  DIR 0=down, 1=up -- amt 0=page, -1 = max
	int i;

	if(amt == -1)                                   // Scroll MAX
	{
		if(dir > 0)                                 // Up MAX
			return (struct sEnqRec *) chainFirst(chnr);
		else                                        // Down MAX
		{
			ptr = (struct sEnqRec *) chainLast(chnr);  // Last entry
			amt = 20;                               // And scroll up 20
			dir = 1;
		}
	}

	if(amt < 1 ||amt > 21)                          // Max Lines is 21
		amt = 21;

	if( dir < 1 && amt == 21 && ptr->next == NULL)  // If at end...
	{
		dir = 1;                                    // Scroll Up 20
		amt = 20;
	}

	if( dir > 0 )                                   // Scroll Up
	{
		for(i=0; i<amt; i++)
		{
			if(ptr->prev != NULL)
				ptr = ptr->prev;
			else
				break;
		}
	}
	else                                            // Scroll Down
	{
		for(i=0; i<amt; i++)
		{
			if(ptr->next != NULL)
				ptr = ptr->next;
			else
				break;
		};
	}

	return ptr;
}


//
//-- Is string all numeric digits
//
static int isNumber(char *str)
{
	while(*str)
	{
		if(!isdigit(*str))
			return 0;
		str++;
	}
	return 1;
}

