/*-----------------------------------------------------------------------------
|  Copyright (c) 2012-2013, Tommy Sprinkle (tommy@tommysprinkle.com)  
|
|  All rights reserved. Redistribution and use in source and binary forms, 
|  with or without modification, are permitted provided that the following 
|  conditions are met:
|
|    * Redistributions of source code must retain the above copyright notice, 
|      this list of conditions and the following disclaimer. 
|    * Redistributions in binary form must reproduce the above copyright  
|      notice, this list of conditions and the following disclaimer in the  
|      documentation and/or other materials provided with the distribution. 
|    * Neither the name of the author nor the names of its contributors may  
|      be used toendorse or promote products derived from this software  
|      without specific prior written permission. 
|
|   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS  
|   "AS IS" AND ANYEXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
|   LIMITED TO, THE IMPLIED WARRANTIESOF MERCHANTABILITY AND FITNESS  
|   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENTSHALL THE 
|   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
|   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
|   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
|   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
|   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
|   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
|   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
|   OF THE POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fss.h"
#include "chainr.h"

void jobInfo(CHAINR jobc, int fts);
static char * fmtCpuTime(char *wk, int cputime);
static char * fmtCpuTimeShort(char *wk, int cputime);
static int isNumber(char *str);
static int procAID(void);
static int cmdShift(char * cmd, char *buf);

#define DASHES "---------------------------------"

struct sJobrec                                      // Job Record - One For Each Address Space
{
	struct sJobrec *next;                           // Pointer to next record
	struct sJobrec *prev;                           // Pointer to previous record

	int  asid;                                      // Binary ASID (record key)
	int  dprty;                                     // Dispatching Priority
	int  disp_pos;                                  // Dispatching Position
	int  cpu_time;                                  // CPU Time Used
	int  srb_time;                                  // SRB Time Used
	int  cpu_delta;                                 // CPU Time Used Since Last Sample
	int  srb_delta;                                 // SRB Time Used Since Last Sample
	int  domain;                                    // Doman
	char jobname[8];                                // Jobname
	char stepname[8];                               // Stepname
	char procname[8];                               // Procname
	char flag[5];                                   // Flags
};



int main(int argc, char ** argv)
{
	CHAINR  jobc;
	struct sJobrec *rec;
	struct sJobrec *scrollPtr;
	int     i;
	int     color;
	int     exit_rc;

	char wk1[12];
	char fld[12];
	char cmd[80];
	char wk[80];

	fssInit();                                      // Initialize Full Screen Environment
#include "sjobpnl.pnc"

	fssSetCursor("COMMAND");                        // Default Cursor Position
	fssSetField("MSG1",DASHES);                     // Clear Error Msg
	fssSetColor("MSG1",0);                          // Default Color

	jobc = chainNewInt(sizeof(struct sJobrec));     // Allocte Chain Storage

	jobInfo(jobc, 1);                               // Get ASID Info (First Time)

	do                                              // Terminal I/O Loop
	{

		jobInfo(jobc, 0);                           // Update ASID Info

		rec = (struct sJobrec *) chainFirst(jobc);  // Start at top

		i = 1;
		while(i < 22 && rec != NULL)                // Max 22 Records to Display
		{
			color = fssBLUE;                        // Default Color Is BLUE
			if(!memcmp(rec->flag, "IN ",3) || !memcmp(rec->flag, "NSW", 3))
			{
				color = fssYELLOW;                  // Non-Swap ASID is YELLOW
				if(rec->cpu_delta > 0 || rec->srb_delta > 0)
					color = fssRED;                 // Using CPU or SRB is RED
			}

			sprintf(fld,"AS%2.2d",i);               // ASID Number
			sprintf(wk1,"%4.4X",rec->asid);
			fssSetField(fld,wk1);
			fssSetColor(fld,color);

			sprintf(fld,"JOB%2.2d",i);              // JOB Name
			fssSetField(fld,rec->jobname);
			fssSetColor(fld,color);

			sprintf(fld,"STEP%2.2",i);              // STEP Name
			fssSetField(fld,rec->stepname);
			fssSetColor(fld,color);

			sprintf(fld,"PROC%2.2d",i);             // PROC Name
			fssSetField(fld,rec->procname);
			fssSetColor(fld,color);

			sprintf(fld,"FLG%2.2d",i);              // Flag
			fssSetField(fld,rec->flag);
			fssSetColor(fld,color);

			sprintf(fld,"DP%2.2d",i);               // Dispatching Priority
			sprintf(wk1,"%3.3d",rec->dprty);
			fssSetField(fld, wk1);
			fssSetColor(fld,color);

			sprintf(fld,"DMN%2.2d",i);              // Domain
			sprintf(wk1,"%2.2d",rec->domain);
			fssSetField(fld, wk1);
			fssSetColor(fld,color);

			sprintf(fld,"CPU%2.2d",i);              // CPU Used
			fssSetField(fld, fmtCpuTime(wk1,rec->cpu_time));
			fssSetColor(fld,color);

			sprintf(fld,"SRB%2.2d",i);              // SRB Used
			fssSetField(fld, fmtCpuTime(wk1,rec->srb_time));
			fssSetColor(fld,color);

			sprintf(fld,"DCPU%2.2d",i);             // Delta CPU
			if(rec->cpu_delta > 0)
			   fssSetField(fld, fmtCpuTimeShort(wk1,rec->cpu_delta));
			else
				fssSetField(fld,"");
			fssSetColor(fld,color);

			sprintf(fld,"DSRB%2.2d",i);             // Delta SRB
			if(rec->srb_delta > 0)
			    fssSetField(fld, fmtCpuTimeShort(wk1,rec->srb_delta));
			else
				fssSetField(fld,"");
			fssSetColor(fld,color);

			i++;
			rec = rec->next;                        // Next Record
		}

		fssRefresh();                               // Update Screen
		fssSetCursor("COMMAND");                    // Default Cursor Position
		fssSetField("MSG1",DASHES);                 // Default Error Msg
	    fssSetColor("MSG1",0);                      // Default Color

		procAID();                                  // Process AID
		strcpy(cmd, fssGetField("COMMAND"));        // Copy Command
		if(cmdShift(cmd, wk))                       // Shift off first token
		{
			if(!strcmp("END",wk))                   // END
			{
				if(cmdShift(cmd,wk))                // No operands allowed
				{
					fssSetField("MSG1", "Invalid Operand for END");
					fssSetColor("MSG1",fssYELLOW);
					fssSetCursor("COMMAND");
				}
				else
				{
					fssReset();
					exit_rc = 4;                    // Set RC=4 and exit
					break;
				}

			}
			if(!strcmp("EXIT",wk))                  // EXIT
			{
				fssReset();
				exit_rc = 8;                        // Set RC=8 and exit
				break;
			}
			else if(!strcmp("DOWN",wk))            //Scroll Down
			{
				if(cmdShift(cmd,wk))
				{
					strupr(wk);
					if(!strcmp(wk,"MAX") || !strcmp(wk,"M"))  // Scroll Max
					{
						//scrollPtr = scroll(ucbchain, scrollPtr, 0, -1);  // Scroll Not Implemented
						fssSetField("COMMAND","");
					}
					else if(!isNumber(wk))          // Number of lines to scroll
					{
						fssSetField("MSG1", "Invalid Scroll Amount");
						fssSetColor("MSG1",fssYELLOW);
						fssSetCursor("COMMAND");
					}
					else
					{
						//scrollPtr = scroll(ucbchain, scrollPtr, 0, atoi(wk));  // Scroll Not Implemented
						fssSetField("COMMAND","");
					}
				}
				else                                // Default - One Page
				{
					//scrollPtr = scroll(ucbchain, scrollPtr, 0, 0);Pointer
					fssSetField("COMMAND","");
				}
			}
			else if(!strcmp("UP",wk))               // UP
			{
				strupr(wk);
				if(cmdShift(cmd,wk))
				{
					if(!strcmp(wk,"MAX") || !strcmp(wk,"M"))   // Up MAX
					{
						//scrollPtr = scroll(ucbchain, scrollPtr, 1, -1);  // Scroll Not Implemented
						fssSetField("COMMAND","");
					}
					else if(!isNumber(wk))          // Number of lines to scroll
					{
						fssSetField("MSG1", "Invalid Scroll Amount");
						fssSetColor("MSG1",fssYELLOW);
						fssSetCursor("COMMAND");
					}
					else
					{
						//scrollPtr = scroll(ucbchain, scrollPtr, 1, atoi(wk));  // Scroll Not Implemented
						fssSetField("COMMAND","");
					}
				}
				else                                // Default - Scroll One Page
				{
					//scrollPtr = scroll(ucbchain, scrollPtr, 1, 0);  // Scroll Not Implemented
					fssSetField("COMMAND","");
				}
			}
			else if(!strcmp("TOP",wk))              // Scroll to TOP
			{
				//scrollPtr = scroll(ucbchain, scrollPtr, 1, -1);  // Scroll Not Implemented
				fssSetField("COMMAND","");
			}
			else if(!strcmp("BOT",wk) || !strcmp("BOTTOM",wk))     // Scroll to Bottom
			{
				//scrollPtr = scroll(ucbchain, scrollPtr, 0, -1);  // Scroll Not Implemented
				fssSetField("COMMAND","");
			}	
			else                                    // Anything Else - Invalid Command
			{
				fssSetField("MSG1","Invalid Command");
				fssSetColor("MSG1",fssYELLOW);
				fssSetCursor("COMMAND");
			}
		}
	} while(1);

	chainFree(jobc);                                // Free Chain Storage

	fssTerm();                                      // Terminate Full Screen Environment

	return exit_rc;
}


//
// Get Info About Each Address Space
//
void jobInfo(CHAINR jobc, int fts)
{
	struct sJobrec *rec;                            // ASID Record Pointer
	unsigned char  *CVT;                            // CVT Address
	unsigned char  *ASVT;                           // ASVT Address
	unsigned char  *SLOT;                           // Address Space Slot
	unsigned char  *ASCB;                           // ASCB Address
	unsigned char  *CSCB;                           // CSCB Address
	unsigned char  *OUCB;                           // OUCB Address
	unsigned char  *TSB;                            // TXB Address
	unsigned char  *nxtCSCB;                        // CSCB Fwd Chain Ptr
	int             asid_count;
	int             t1;
	int             t2;


	int             asid;                           // ASID
	int             dprty;                          // Dispatching Priority
	int             disp_pos;                       // Dispatching Position
	int             cpu_time;                       // CPU Time
	int             srb_time;                       // SRB Time
	int             domain;                         // Domain

	char            jobname[8];
	char            stepname[8];
	char            procname[8];
	char            flag[5];

	CVT  = *( (unsigned char **) (0x10));           // Point to CVT
	ASVT = *( (unsigned char **) (CVT+0x22C));      // Point to ASVT
	SLOT = ASVT+0x20C;                              // First Address Space Slot
	asid_count = *( (int *) (ASVT+0x204));          // Number of ASID Slots

	while(asid_count > 0)                           // Loop through Slots
	{
		if( ((*SLOT) & 0x80) != 0x80)               // If In-Use
		{
			ASCB = *( (unsigned char **) (SLOT));        // ASCB
			CSCB = *( (unsigned char **) (ASCB+0x38));   // CSCB
			OUCB = *( (unsigned char **) (ASCB+0x90));   // OUCB
			TSB  = *( (unsigned char **) (ASCB+0x3C));  

			asid = ((*(ASCB+0x24)) << 8) + *(ASCB+0x25);    // Get ASID
			dprty = *(ASCB+0x2B);                           // Get Disptach Priority
			disp_pos = ((*(ASCB+26)) << 8) + *(ASCB+0x27);  //Get Dispatch Position
			t1 = *( (unsigned int *) (ASCB+0x40));          // CPU Time Double Word
			t2 = *( (unsigned int *) (ASCB+0x44));          // SRDL 22
			cpu_time = ((t1 << 10) & 0xFFFFFC00) + ((t2 >> 22) & 0x3FF);
			cpu_time /= 10;                                 // Divide By 10

			t1 = *( (unsigned int *) (ASCB+0xC8));          // SRB Time Double Word
			t2 = *( (unsigned int *) (ASCB+0xCC));          // SRDL 22
			srb_time = ((t1 << 10) & 0xFFFFFC00) + ((t2 >> 22) & 0x3FF);
			srb_time /= 10;                                 // Divide by 10

			if( *(CSCB+0x1C) == 0x01 )              // TSO User
			{
				memcpy(jobname,  CSCB+0x08,  8);    // Job Name
				memcpy(stepname, CSCB+0x10,  8);    // Step Name
				memcpy(procname, "        ", 8);    // No Proc Name
			}
			else if( *(CSCB+0x1C) == 0x03 )         // INIT
			{
				nxtCSCB = CSCB;                     // Look for executing JOB CSCB
				while(1)
				{
					nxtCSCB = *( (unsigned char **) (nxtCSCB));  // Scan forward in CSCB Chain
					if(nxtCSCB == NULL || nxtCSCB == CSCB)       // End of Chain, nothing found
					{
						memcpy(jobname,  CSCB+0x08,  8);         // Use as Job Name
						memcpy(stepname, "        ", 8);
						memcpy(stepname, "        ", 8);
						break;
					}
					if(asid == (*(nxtCSCB+0x1E) << 8) + *(nxtCSCB+0x1F))  // Same ASID as INIT
					{
						memcpy(jobname,  nxtCSCB+0x08, 8);       // Job Name
						memcpy(stepname, nxtCSCB+0x40, 8);       // Step Name
						memcpy(procname, nxtCSCB+0x20, 8);       // Proc Name
						break;
					}
				}
			}
			else                                    // Not TSO or INIT (STC)
			{
				memcpy(jobname,  CSCB+0x10, 8);     // Job Name
				memcpy(stepname, CSCB+0x08, 8);     // Step Name
				memcpy(procname, CSCB+0x20, 8);     // Proc Name
				if(asid == 1)                       // If ASID = 1
				{
					memcpy(jobname,  "MASTER  ",8);
					memcpy(stepname, "        ", 8);
					memcpy(procname, "        ", 8);
				}
			}

			domain = *(OUCB+0x24);                  // Save Domain

			memcpy(flag,"     ",5);                 // Clear Flags
			if( (*(OUCB+0x11) & 0x80) == 0x80)      // Non-Swap
				memcpy(flag,"NSW",3);
			else if( *(OUCB+0x25) == 0x00 )         // Swapped IN
				memcpy(flag,"IN ",3);
			else
			{
				if( (*(OUCB+0x10) & 0xE0) == 0x00)  // Swapped OUT
					memcpy(flag,"OUT",3);
				else if( (*(OUCB+0x10) & 0x80) == 0x80) // Transitioning Out
					memcpy(flag,"GOO",3);
				else if( (*(OUCB+0x10) & 0x20) == 0x20) // Transitioning In
					memcpy(flag,"GOB",3);
				else                                    // Transitioning Between States
					memcpy(flag,"GOI",3);
			}

			if( (*(OUCB+0x12) & 0x04) == 0x04)          // AUX Storage Swap
				memcpy(flag+3,"A ",2);
			else if( (*(OUCB+0x12) & 0x01) == 0x01)     // Fixed Storage Swap
				memcpy(flag+3,"S ",2);
			else if( (*(OUCB+0x15) & 0x80) == 0x80)     // Long Wait
				memcpy(flag+3,"LW",2);
			else if( (*(OUCB+0x15) & 0x01) == 0x01)     // Detected Wait
				memcpy(flag+3,"MW",2);
			else if( (*(OUCB+0x15) & 0x10) == 0x10)     // 
				memcpy(flag+3,"CP",2);

			if(fts)                                 // If Frist Time, Create Record
			{
				rec = (struct sJobrec *) chainAdd(jobc, asid);
				rec->dprty     = dprty;
				rec->disp_pos  = disp_pos;
				rec->cpu_time  = cpu_time;
				rec->srb_time  = srb_time;
				rec->cpu_delta = 0;
				rec->srb_delta = 0;
				rec->domain    = domain;
				memcpy(rec->jobname,  jobname,  8);
				memcpy(rec->stepname, stepname, 8);
				memcpy(rec->procname, procname, 8);
				memcpy(rec->flag,     flag,     5);
			}
			else                                    // Not First Time    
			{
				rec = (struct sJobrec *) chainFind(jobc, asid);  // See if record Exists
				if(rec == NULL)                                  // No - Create it
				{
					rec = (struct sJobrec *) chainAdd(jobc, asid);
					rec->dprty     = dprty;
					rec->disp_pos  = disp_pos;
					rec->cpu_time  = cpu_time;
					rec->srb_time  = srb_time;
					rec->cpu_delta = 0;
					rec->srb_delta = 0;
					rec->domain    = domain;
					memcpy(rec->jobname,  jobname,  8);
					memcpy(rec->stepname, stepname, 8);
					memcpy(rec->procname, procname, 8);
					memcpy(rec->flag,     flag,     5);
				}
				else                                // Record Exists - Update It
				{
					if(cpu_time > rec->cpu_time)
						rec->cpu_delta = cpu_time - rec->cpu_time;
					else
						rec->cpu_delta = 0;

					if(srb_time > rec->srb_time)
						rec->srb_delta = srb_time - rec->srb_time;
					else
						rec->srb_delta = 0;

					rec->cpu_time  = cpu_time;
					rec->srb_time  = srb_time;

					rec->dprty     = dprty;
					rec->disp_pos  = disp_pos;
					rec->domain    = domain;
					memcpy(rec->jobname,  jobname,  8);
					memcpy(rec->stepname, stepname, 8);
					memcpy(rec->procname, procname, 8);
					memcpy(rec->flag,     flag,     5);
				}
			}
		}
		SLOT += 4;                                  // Next ASID Slot
		asid_count--;                               // Decrement ASID Count
	}
}


//
//  Format CPU Time MM.SS.HH
//
static char * fmtCpuTime(char *wk, int cputime)
{
	int hh;
	int ss;
	int mm;


	hh = cputime % 100;
	cputime /= 100;
	ss = cputime % 60;
	mm = cputime / 60;

	sprintf(wk,"%2.1d.%2.2d.%2.2d",mm,ss,hh);
	return wk;
}

//
// Format Delta CPU Time SS.HH
//
static char * fmtCpuTimeShort(char *wk, int cputime)
{
	int hh;
	int ss;
	int mm;


	hh = cputime % 100;
	cputime /= 100;
	ss = cputime % 60;
	mm = cputime / 60;

	sprintf(wk,"%2.1d.%2.2d",ss,hh);
	return wk;
}


//
//-- shift off a command token
//
static int cmdShift(char * cmd, char *buf)
{
	int   len;
	char *p;
	char *p1;

	p = cmd;                                      // start of command string
	while(*p == ' ')                              // skip over blanks
		p++;

	if(!*p)                                       // if end of string
		return 0;

	p1 = p;                                       // remember start of token
	while(*p && *p != ' ')                        // find end of token
		p++;

	len = p - p1;                                 // token length
	memcpy(buf, p1, len);                         // copy token
	*(buf+len) = '\0';

	strcpy(cmd, p);                               // shift over past removed token
	return 1;
}

//
//-- process AID
//
static int procAID(void)
{
	int    aid;
	char   wk[80];

	wk[0] = '\0';
	aid = fssGetAID();                            // Get AID value
	if( aid == fssPFK03 || aid == fssPFK15)       // PFK 3 or 15
		strcpy(wk, "END ");
	if( aid == fssPFK04 || aid == fssPFK16)       // PFK 4 or 16
		strcpy(wk, "EXIT ");
	else if(aid == fssPFK08 || aid == fssPFK20)   // PFK 8 or 20
		strcpy(wk, "DOWN ");
	else if(aid == fssPFK07 || aid == fssPFK19)   // PFK 7 or 19
		strcpy(wk, "UP ");

	strcat(wk, fssGetField("COMMAND"));           // Append contents of COMMAND field
	fssSetField("COMMAND", strupr(wk));           // Copy back to COMMAND

	return 0;
}


//
//-- Is string all numeric digits
//
static int isNumber(char *str)
{
	while(*str)
	{
		if(!isdigit(*str))
			return 0;
		str++;
	}
	return 1;
}

