/*-----------------------------------------------------------------------------
|  Copyright (c) 2012-2013, Tommy Sprinkle (tommy@tommysprinkle.com)  
|
|  All rights reserved. Redistribution and use in source and binary forms, 
|  with or without modification, are permitted provided that the following 
|  conditions are met:
|
|    * Redistributions of source code must retain the above copyright notice, 
|      this list of conditions and the following disclaimer. 
|    * Redistributions in binary form must reproduce the above copyright  
|      notice, this list of conditions and the following disclaimer in the  
|      documentation and/or other materials provided with the distribution. 
|    * Neither the name of the author nor the names of its contributors may  
|      be used toendorse or promote products derived from this software  
|      without specific prior written permission. 
|
|   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS  
|   "AS IS" AND ANYEXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
|   LIMITED TO, THE IMPLIED WARRANTIESOF MERCHANTABILITY AND FITNESS  
|   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENTSHALL THE 
|   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
|   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
|   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
|   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
|   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
|   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
|   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
|   OF THE POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fss.h"
#include "chainr.h"

#define DASHES "---------------------------------"

static int CSAinfo(void);
static int memMap(void);
static int cmdShift(char * cmd, char *buf);
static int procAID(void);
static int isNumber(char *str);

int main(int argc, char ** argv)
{
	char cmd[80];
	char wk[80];

	int  exit_rc;


	fssInit();

#include "srsmpnl1.pnc"

	fssSetCursor("COMMAND");						// Default Cursor Location
	fssSetField("MSG1",DASHES);						// Clear Error Msg
	fssSetColor("MSG1",0);							// Default Color

	while(1)										// Main terminal loop
	{
		memMap();									// Build Memory Map
		CSAinfo();									// CSA Info

		fssSetCursor("COMMAND");					// Default Cursor Location
		fssSetField("MSG1",DASHES);					// Clear Error Msg
		fssSetColor("MSG1",0);						// Defualt Color

		fssRefresh();								// Display Screen

		procAID();                                  // Process AID
		strcpy(cmd, fssGetField("COMMAND"));        // Copy Command
		if(cmdShift(cmd, wk))                       // Shift off first token
		{
			if(!strcmp("END",wk))                   // END
			{
				if(cmdShift(cmd,wk))                // No operands allowed
				{
					fssSetField("MSG1", "Invalid Operand for END");
					fssSetColor("MSG1",fssYELLOW);
					fssSetCursor("COMMAND");
				}
				else
				{
					fssReset();
					exit_rc = 4;                    // Set RC=4 and exit
					break;
				}

			}
			if(!strcmp("EXIT",wk))                  // EXIT
			{
				fssReset();
				exit_rc = 8;                        // Set RC=8 and exit
				break;
			}
			else                                    // Anything Else - Invalid Command
			{
				fssSetField("MSG1","Invalid Command");
				fssSetColor("MSG1",fssYELLOW);
				fssSetCursor("COMMAND");
			}
		}
	}

	fssTerm();										// Terminate Fullscreen Environment

	return exit_rc;
}


//
// Build Memory Map
//
static int memMap(void)
{
	unsigned char *CVT;
	unsigned char *PVT;
	unsigned char *GDA;
	unsigned char *CSA_PQE;
	unsigned char *SQA_PQE;
	unsigned char *SYSR_PQE;
	unsigned char *SYSV_PQE;
	unsigned char *SYS_PQE;
	unsigned char *DQE;
	unsigned char *ASCB;
	unsigned char *LDA;

	int			   commonStart;
	int			   commonEnd;
	int			   commonSize;

	int			   SQAstart;
	int			   SQAend;
	int			   SQAsize;

	int            PLPAstart;
	int            PLPAend;
	int            PLPAsize;

	int			   CSAstart;
	int			   CSAend;
	int			   CSAsize;

	int			   privateStart;
	int			   privateEnd;
	int			   privateSize;

	int			   VVRegionStart;
	int			   VVRegionEnd;
	int			   VVRegionSize;

	int			   VRRegionStart;
	int			   VRRegionEnd;
	int			   VRRegionSize;

	int			   SYSRegionStart;
	int			   SYSRegionEnd;
	int			   SYSRegionSize;

	int			   NucleusStart;
	int			   NucleusEnd;
	int			   NucleusSize;

	char           wk1[16];
	char           wk2[16];
	char           wk3[16];

	CVT = *((unsigned char **) (0x10));
	PVT = *((unsigned char **) (CVT+0x164));
	GDA = *((unsigned char **) (CVT+0x230));

	CSA_PQE = *((unsigned char **) (GDA+0x8));		// CSA PQE

	commonStart = *((int *) (CSA_PQE+0x18));		// Start of Common Area
	commonEnd   = *((int *) (CVT+0xA4));			// End of Common Area (Top of Memory)
	commonSize  = commonEnd - commonStart +1;		// Common Size

	SQA_PQE = *((unsigned char **) (GDA+0x18));		// SQA PQE
	DQE     = *((unsigned char **) (SQA_PQE+0x04));

	SQAstart = *((int *) (DQE+0x08));				// Start of SQA
	SQAsize  = *((int *) (DQE+0x0C));				// SQA Size
	SQAend   = SQAstart + SQAsize -1;				// Calc End Address

	PLPAend = *((int *) (DQE+0x08));				// PLPA End (SQA start -1)
	PLPAend--;	
	PLPAstart = ((*(CVT+0x169) << 16) + (*(CVT+0x16A) << 8) + *(CVT+0x16B));
	PLPAsize  = PLPAend - PLPAstart +1;				// Calc Size

	CSAstart = *((int *) (CSA_PQE+0x18));			// CSA Start
	CSAsize  = *((int *) (CSA_PQE+0x14));			// CSA Size
	CSAend   = CSAstart + CSAsize -1;				// Calc End Address

	privateStart = *((int *) (GDA+0x10));			// Private Start
	privateSize   = *((int *) (GDA+0x14));			// Private Size
	privateEnd  = privateStart + privateSize;		// Calc End Address

	ASCB = *((unsigned char **) (0x224));			// PSAAOLD = Current ASCB 
	LDA  = *((unsigned char **) (ASCB+0x30));
	SYSV_PQE = *((unsigned char **) (LDA+0x08));

	VVRegionStart = *((int *) (SYSV_PQE+0x18));		// V=V Region
	VVRegionSize  = *((int *) (SYSV_PQE+0x14));
	VVRegionEnd   = VVRegionStart + VVRegionSize -1;

	SYSR_PQE = *((unsigned char **) (GDA+0x0C));	// V=R Region
	VRRegionStart = *((int *) (SYSR_PQE+0x18));
	VRRegionSize  = *((int *) (SYSR_PQE+0x14));
	VRRegionEnd   = VRRegionStart + VRRegionSize -1;

	SYS_PQE = *((unsigned char **) (LDA+0x5C4));	// System Region
	SYSRegionStart = *((int *) (SYS_PQE+0x18));
	SYSRegionSize  = *((int *) (SYS_PQE+0x14));
	SYSRegionEnd   = SYSRegionStart + SYSRegionSize -1;

	NucleusStart = 0;								// Nucleus starts at beginning
	NucleusEnd   = *((int *) (CVT+0x80));			// Nucleus End Address
	NucleusEnd--;
	NucleusSize  = NucleusEnd +1;

	sprintf(wk1, "%6.6X", commonStart);
	sprintf(wk2, "%6.6X", commonEnd);
	sprintf(wk3, "%4dK",  commonSize/1024);
	fssSetField("COMST",  wk1);
	fssSetField("COMEND", wk2);
	fssSetField("COMSZ",  wk3);

	sprintf(wk1, "%6.6X", SQAstart);
	sprintf(wk2, "%6.6X", SQAend);
	sprintf(wk3, "%4dK",  SQAsize/1024);
	fssSetField("SQAST",  wk1);
	fssSetField("SQAEND", wk2);
	fssSetField("SQASZ",  wk3);

	sprintf(wk1, "%6.6X", PLPAstart);
	sprintf(wk2, "%6.6X", PLPAend);
	sprintf(wk3, "%4dK",  PLPAsize/1024);
	fssSetField("PLPST",  wk1);
	fssSetField("PLPEND", wk2);
	fssSetField("PLPSZ",  wk3);

	sprintf(wk1, "%6.6X", CSAstart);
	sprintf(wk2, "%6.6X", CSAend);
	sprintf(wk3, "%4dK",  CSAsize/1024);
	fssSetField("CSAST",  wk1);
	fssSetField("CSAEND", wk2);
	fssSetField("CSASZ",  wk3);

	sprintf(wk1, "%6.6X", privateStart);
	sprintf(wk2, "%6.6X", privateEnd);
	sprintf(wk3, "%4dK",  privateSize/1024);
	fssSetField("PRIST",  wk1);
	fssSetField("PRIEND", wk2);
	fssSetField("PRISZ",  wk3);

	sprintf(wk1, "%6.6X", VVRegionStart);
	sprintf(wk2, "%6.6X", VVRegionEnd);
	sprintf(wk3, "%4dK",  VVRegionSize/1024);
	fssSetField("VVST",  wk1);
	fssSetField("VVEND", wk2);
	fssSetField("VVSZ",  wk3);

	sprintf(wk1, "%6.6X", VRRegionStart);
	sprintf(wk2, "%6.6X", VRRegionEnd);
	sprintf(wk3, "%4dK",  VRRegionSize/1024);
	fssSetField("VRST",  wk1);
	fssSetField("VREND", wk2);
	fssSetField("VRSZ",  wk3);

	sprintf(wk1, "%6.6X", SYSRegionStart);
	sprintf(wk2, "%6.6X", SYSRegionEnd);
	sprintf(wk3, "%4dK",  SYSRegionSize/1024);
	fssSetField("SYSST",  wk1);
	fssSetField("SYSEND", wk2);
	fssSetField("SYSSZ",  wk3);

	sprintf(wk1, "%6.6X", NucleusStart);
	sprintf(wk2, "%6.6X", NucleusEnd);
	sprintf(wk3, "%4dK",  NucleusSize/1024);
	fssSetField("NUCST",  wk1);
	fssSetField("NUCEND", wk2);
	fssSetField("NUCSZ",  wk3);

	return 0;

}
static int CSAinfo(void)
{
	unsigned char *CVT;
	unsigned char *PVT;
	unsigned char *GDA;
	unsigned char *PQE;
	unsigned char *FBQE;

	int            freeCSA;
	int            sizeCSA;
	int            freeSQA;
	int            availSQA;

	char           wk[16];


	CVT = *((unsigned char **) (0x10));
	PVT = *((unsigned char **) (CVT+0x164));
	GDA = *((unsigned char **) (CVT+0x230));

	PQE = *((unsigned char **) (GDA+0x8));		// CSA PQE

	FBQE = *((unsigned char **) (PQE));			// 1st FBQE for CSA
	freeCSA = 0;

	while(PQE != FBQE)
	{
		freeCSA += *((int *) (FBQE+8));			// Add Free Space
		FBQE = *((unsigned char **) (FBQE));	// Next FBQE
	}
	

	freeSQA = *((int *) (GDA+0x1C));			// SQA Available
	freeSQA -= freeCSA;							// Subtract CSA
	if(freeSQA < 0)
		freeSQA = 0;
	freeSQA = freeSQA >> 10;					// Divide by 1K

	availSQA = *((int *) (GDA+0x1c));			// SQA Available
	availSQA = availSQA >> 10;					// Divide by 1K

	freeCSA  = freeCSA  >> 10;					// Divide by 1K	

	sprintf(wk, "%4dK", freeCSA);
	fssSetField("FREECSA", wk);

	sprintf(wk, "%4dK", freeSQA);
	fssSetField("FREESQA", wk);

	sprintf(wk, "%4dK", availSQA);
	fssSetField("AVALSQA", wk);

	return 0;
}



//
//-- shift off a command token
//
static int cmdShift(char * cmd, char *buf)
{
	int   len;
	char *p;
	char *p1;

	p = cmd;                                      // start of command string
	while(*p == ' ')                              // skip over blanks
		p++;

	if(!*p)                                       // if end of string
		return 0;

	p1 = p;                                       // remember start of token
	while(*p && *p != ' ')                        // find end of token
		p++;

	len = p - p1;                                 // token length
	memcpy(buf, p1, len);                         // copy token
	*(buf+len) = '\0';

	strcpy(cmd, p);                               // shift over past removed token
	return 1;
}

//
//-- process AID
//
static int procAID(void)
{
	int    aid;
	char   wk[80];

	wk[0] = '\0';
	aid = fssGetAID();                            // Get AID value
	if( aid == fssPFK03 || aid == fssPFK15)       // PFK 3 or 15
		strcpy(wk, "END ");
	if( aid == fssPFK04 || aid == fssPFK16)       // PFK 4 or 16
		strcpy(wk, "EXIT ");
	else if(aid == fssPFK08 || aid == fssPFK20)   // PFK 8 or 20
		strcpy(wk, "DOWN ");
	else if(aid == fssPFK07 || aid == fssPFK19)   // PFK 7 or 19
		strcpy(wk, "UP ");

	strcat(wk, fssGetField("COMMAND"));           // Append contents of COMMAND field
	fssSetField("COMMAND", strupr(wk));           // Copy back to COMMAND

	return 0;
}


//
//-- Is string all numeric digits
//
static int isNumber(char *str)
{
	while(*str)
	{
		if(!isdigit(*str))
			return 0;
		str++;
	}
	return 1;
}

