/*-----------------------------------------------------------------------------
|  Copyright (c) 2012-2013, Tommy Sprinkle (tommy@tommysprinkle.com)  
|
|  All rights reserved. Redistribution and use in source and binary forms, 
|  with or without modification, are permitted provided that the following 
|  conditions are met:
|
|    * Redistributions of source code must retain the above copyright notice, 
|      this list of conditions and the following disclaimer. 
|    * Redistributions in binary form must reproduce the above copyright  
|      notice, this list of conditions and the following disclaimer in the  
|      documentation and/or other materials provided with the distribution. 
|    * Neither the name of the author nor the names of its contributors may  
|      be used toendorse or promote products derived from this software  
|      without specific prior written permission. 
|
|   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS  
|   "AS IS" AND ANYEXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
|   LIMITED TO, THE IMPLIED WARRANTIESOF MERCHANTABILITY AND FITNESS  
|   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENTSHALL THE 
|   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
|   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
|   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
|   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
|   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
|   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
|   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
|   OF THE POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <process.h>
#include "fss.h"

static int getSystemValues(void);
static int getEnvValues(void);
static int cmdShift(char * cmd, char *buf);
static int procAID(void);
static int  getMasterCat(void);
static int getIPL(void);

#define DASHES "---------------------------------"

int main(int argc, char ** argv)
{
	int  aid;
	char cmd[80];
	char wk[80];


	fssInit();

#include "sstatpnl.pnc"

	fssSetField("MSG1",DASHES);                  // Clear Error Msg
	fssSetColor("MSG1",0);                       // Default Color
	fssSetCursor("COMMAND");                     // Default Cursor Position

	while(1)                                     // Main Terminal I/O Loop
	{
		getSystemValues();                       // Get System Info
		getEnvValues();                          // Get Environment Info
		getMasterCat();							 // Get Master Catalog DSNAME & VolSer
		getIPL();								 // Get IPL Info

		fssRefresh();                            // Display Screen

		fssSetField("MSG1",DASHES);              // Clear Error Msg
		fssSetColor("MSG1",0);                   // Default Color
		fssSetCursor("COMMAND");                 // Cursor Position

		 procAID();                              // Process AID 

	   strcpy(cmd, fssGetField("COMMAND"));      // copy contents of Command field
	   if(cmdShift(cmd, wk))                     // shift off first token
       {
          if(!strcmp("END",wk))                  // is it END
          {
             if(cmdShift(cmd,wk))                // No operands allowed
             {
                fssSetField("MSG1", "Invalid Operand for END");
				fssSetColor("MSG1",fssYELLOW);   // Set color for Err Msg
                fssSetCursor("COMMAND");         // Set cursor
             }
             else
             {
                break;                           // Exit
             }

          }
          if(!strcmp("EXIT",wk))                 // Is Command "EXIT"
          {
             break;                              // Exit
          }
		  else if(!strcmp("1",wk))               // Option = 1
		  {
			  system("JOBINFO");                 // Call Address Space Info
		      fssSetField("COMMAND","");         // Clear Command
		  }
		  else if(!strcmp("2",wk))               // Option = 2
		  {
			  system("UCBSCAN");                 // Device Info
		      fssSetField("COMMAND","");         // Clear Command
		  }
		  else if(!strcmp("3",wk))               // Option = 3
		  {
			  system("SWAPINFO");                // Paging & Swap Info
		      fssSetField("COMMAND","");         // Clear Command
		  }
		  else if(!strcmp("4",wk))               // Option = 4
		  {
			  system("ENQINFO");                // Paging & Swap Info
		      fssSetField("COMMAND","");         // Clear Command
		  }
		  else if(!strcmp("5",wk))               // Option = 5
		  {
			  system("RSMINFO");                // Paging & Swap Info
		      fssSetField("COMMAND","");         // Clear Command
		  }
		  else
          {
             fssSetField("MSG1","Invalid Command");
			 fssSetColor("MSG1",fssYELLOW);      // Set highlight color
             fssSetCursor("COMMAND");           
          }
       }


	}

	fssTerm();

	return 0;
}

//
//  Get Environmental Information
//
static int getEnvValues(void)
{
	time_t     timer;
	struct tm *curTime;
	
	char       userid[9];
	char       date[12];
	char       jdate[12];
	char       tod[12];

	time(&timer);                                // Get Current Time
	curTime = localtime(&timer);                 // Format Date & Time
	sprintf(date , "%2.2d/%2.2d/%4.4d",(curTime->tm_mon)+1,curTime->tm_mday,(curTime->tm_year)+1900);
	sprintf(jdate, "%2.2d.%3.3d",(curTime->tm_year)%100,(curTime->tm_yday)+1);
	sprintf(tod  , "%2.2d:%2.2d:%2.2d",curTime->tm_hour,curTime->tm_min,curTime->tm_sec);
	cuserid(userid);                             // Get USERID

	fssSetField("USERID",userid);                // Set Screenn Fields...
	fssSetField("TIME",  tod   );
	fssSetField("DATE",  date  );
	fssSetField("JDATE", jdate );

	return 0;
}


//
// Get Master Catalog Name
//
static int  getMasterCat(void)
{
	unsigned char *CVT;
	unsigned char *AMCBS;
	unsigned char *CAXWA;
	unsigned char *nextCAXWA;
	unsigned char *UCB;

	char           dsname[44];
	char           volser[6];

	CVT = *((unsigned char **) (0x10));				
	AMCBS = *((unsigned char **) (CVT+0x100));
	CAXWA = *((unsigned char **) (AMCBS+0x14));

	nextCAXWA = *((unsigned char **) (CAXWA+0x04));	// Next On Chain

	while(nextCAXWA != NULL)						// Last Entry is Master Cat
	{
		CAXWA = nextCAXWA;
		nextCAXWA = *((unsigned char **) (CAXWA+0x04));
	}

	memcpy(dsname, CAXWA+0x34, 44);					// Master Cat DSN
	UCB = *((unsigned char **) (CAXWA+0x1C));		// Master Cat UCB
	memcpy(volser, UCB+0x1C, 6);					// Copy VOLSER

	fssSetField("CATDSN",dsname);
	//fssSetColor("CATDSN",fssPINK);
	fssSetField("CATVOL",volser);
	//fssSetColor("CATVOL",fssPINK);

	return 0;
}


//
// Get IPL Information
//
static int getIPL(void)
{
	unsigned char *CVT;
	unsigned char *UCB;
	unsigned char *ASMVT;
	unsigned char *CVTEXT2;

	char           sysres[6];
	char           sysipl[3];
	char           clpa[3];
	char           nucid[8];

	CVT     = *((unsigned char **) (0x10));
	UCB     = *((unsigned char **) (CVT+0x30));		// UCB of SYSRES
	ASMVT   = *((unsigned char **) (CVT+0x2C0));
	CVTEXT2 = *((unsigned char **) (CVT+0x148));

	memcpy(sysres, UCB+0x1C, 6);					// IPL VOLSER
	memcpy(sysipl, UCB+0x0D, 3);					// IPL Device Addresss
	memcpy(nucid, "IEANUC00", 8);					// Nuc Module Name
	nucid[7] = *(CVTEXT2+0x04);						// IPL Nuc ID

	if( ((*(ASMVT+1)) & 0x08) == 0x08 )				// TM ASMFLAG2,ASMQUICK
		memcpy(clpa, "No ", 3);						//    CLPA=NO
	else
		memcpy(clpa, "Yes", 3);						//    CLPA=YES

	fssSetField("SYSRES", sysres);
	fssSetField("SYSIPL", sysipl);
	fssSetField("CLPA",   clpa);
	fssSetField("NUCID",  nucid);

	return 0;
}



//
//  Get System Information
//
static int getSystemValues(void)
{
	unsigned char *CVT;
	unsigned char **CVTPTR;
	unsigned char *SMCA;
	unsigned char *PCCA;

	int            iWork;
	char           wk[80];

	CVTPTR = (unsigned char **) 0x10;            // CVT Pointer
	CVT    = *CVTPTR;                            // CVT Address

	iWork  = (*(CVT-6) << 8) + *(CVT-5);         // CPU Model
	sprintf(wk,"%2.2X", iWork);
	fssSetField("CPUMODEL",wk);

	PCCA = *((char **) (CVT+0x2FC));             // Find CPU Serial Number
	PCCA = *((char **) (PCCA));
	sprintf(wk,"%6.6s", (char *) PCCA+6);
	fssSetField("CPUSER", wk);

	sprintf(wk,"%4.4s", (CVT-4));                // OS Release Number
	fssSetField("OSREL", wk);

	iWork = *( (int *) (CVT+0xA4));              // Memory Size
	iWork = iWork/1024;       // K
	if(iWork < 1000)                             // Convert to K
		sprintf(wk,"%d K",iWork);
	else
		sprintf(wk,"%d,%d K",iWork/1000,iWork%1000);
	fssSetField("MEMSIZE",wk);

	SMCA = *( (char **) (CVT+0xC4));             // SMCA - SMF Status

	sprintf(wk,"%4.4s", (SMCA+0x10) );           // System ID
	fssSetField("SID",wk);

	sprintf(wk,"%6.6s", (SMCA+0x18) );           // Primary (Active) SMF Dataset VOLSER
	fssSetField("PSMFVL",wk);

	sprintf(wk,"SYS1.MAN%1.1s", (SMCA+0x23) );   // Primary (Active) SMF Dataset
	fssSetField("PSMFDS", wk);

	sprintf(wk,"%6.6s", (SMCA+0x28) );           // Secondary SMF Dataset Volser
	fssSetField("SSMFVL",wk);

	sprintf(wk,"SYS1.MAN%1.1s", (SMCA+0x33) );   // Secondary SMF Dataset Volser
	fssSetField("SSMFDS",wk);

	return 0;
}

//
//-- Process AID
//
static int procAID(void)
{
   int    aid;
   char   wk[80];

   wk[0] = '\0';                                 // Clear work string
   aid = fssGetAID();                            // Get AID value
   if( aid == fssPFK03 || aid == fssPFK15)       // PFK 3 or 15
      strcpy(wk, "END ");
   if( aid == fssPFK04 || aid == fssPFK16)       // PFK 4 or 16
      strcpy(wk, "EXIT ");
   else if(aid == fssPFK08 || aid == fssPFK20)   // PFK 8 or 20
      strcpy(wk, "DOWN ");
   else if(aid == fssPFK07 || aid == fssPFK19)   // PFK 7 or 19
      strcpy(wk, "UP ");

   strcat(wk, fssGetField("COMMAND"));           // add contenst of Command 
   fssSetField("COMMAND", strupr(wk));           // put back updated Command

   return 0;
}


//
//-- Shift off Command Token
//
static int cmdShift(char * cmd, char *buf)
{
   int   len;
   char *p;
   char *p1;

   p = cmd;                                      // set pointer to beginning of string
   while(*p == ' ')                              // skip over any blanks
      p++;

   if(!*p)                                       // return zero if end of string
      return 0;

   p1 = p;                                       // remember start of token
   while(*p && *p != ' ')                        // scan for end of token
      p++;

   len = p - p1;                                 // calc length of token
   memcpy(buf, p1, len);                         // return token;
   *(buf+len) = '\0';

   strcpy(cmd, p);                               // shift contents over
   return 1;                                     // token returned
}



