/*-----------------------------------------------------------------------------
|  Copyright (c) 2012-2013, Tommy Sprinkle (tommy@tommysprinkle.com)  
|
|  All rights reserved. Redistribution and use in source and binary forms, 
|  with or without modification, are permitted provided that the following 
|  conditions are met:
|
|    * Redistributions of source code must retain the above copyright notice, 
|      this list of conditions and the following disclaimer. 
|    * Redistributions in binary form must reproduce the above copyright  
|      notice, this list of conditions and the following disclaimer in the  
|      documentation and/or other materials provided with the distribution. 
|    * Neither the name of the author nor the names of its contributors may  
|      be used toendorse or promote products derived from this software  
|      without specific prior written permission. 
|
|   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS  
|   "AS IS" AND ANYEXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
|   LIMITED TO, THE IMPLIED WARRANTIESOF MERCHANTABILITY AND FITNESS  
|   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENTSHALL THE 
|   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
|   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
|   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
|   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
|   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
|   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
|   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
|   OF THE POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------*/

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "fss.h"


struct sSwapCounts                                  // Swap Count Data Area
{
	int term_wait_in;
	int term_wait_out;
	int long_wait;
	int aux_stor_shortage;
	int real_stor_shortage;
	int detected_wait;
	int request_swap;
	int enq_exchange;
	int workload_exchange;
	int unilateral_swap;
	char ips[4];
};

struct sAsmCounts                                   // Aux Storage Manager Counts Data Area
{
	int  avail;
	int  total;
	int  vio;
	int  adspc;
	int  bad;
};

static void getSwapCounts(struct sSwapCounts *sc);
static void getPageDS(int *);
static void getSwapDS(int *);
static void getASMCounts(struct sAsmCounts *asm);
char *      itoac(char *buf, int n);

#define DASHES "---------------------------------"

int main(int argc, char ** argv)
{
	struct sSwapCounts sc;
	struct sAsmCounts  asm;
	int    linect;
	char   wk[16];

	fssInit();                                      // Initialize Full Screen Environment
#include "sswappnl.pnc"

	fssSetCursor("COMMAND");                        // Default Cursor Position
	fssSetField("MSG1",DASHES);						// Clear Error Msg
	fssSetColor("MSG1",0);							// Default Color

	while(1)										// Main Terminal Loop
	{
		getSwapCounts(&sc);							// Get Swap Counts
		fssSetField("IPS",sc.ips);	                // Format Data To Screen
		itoac(wk, sc.term_wait_in);
		fssSetField("TRMWTI",itoac(wk,sc.term_wait_in));
		fssSetField("TRMWTO",itoac(wk,sc.term_wait_out));
		fssSetField("LONGWT",itoac(wk,sc.long_wait));
		fssSetField("AUXSS", itoac(wk,sc.aux_stor_shortage));
		fssSetField("REALSS",itoac(wk,sc.real_stor_shortage));	
		fssSetField("DETWT", itoac(wk,sc.detected_wait));
		fssSetField("REQSW", itoac(wk,sc.request_swap));
		fssSetField("ENQEXC",itoac(wk,sc.enq_exchange));
		fssSetField("WKLEXC",itoac(wk,sc.workload_exchange));
		fssSetField("UNISWP",itoac(wk,sc.unilateral_swap));

		getASMCounts(&asm);							// Get ASM Counts
		fssSetField("AVAIL",itoac(wk,asm.avail));	// Format to Screen
		fssSetField("TOTAL",itoac(wk,asm.total));
		fssSetField("VIO"  ,itoac(wk,asm.vio));
		fssSetField("ADSPC",itoac(wk,asm.adspc));
		fssSetField("BAD"  ,itoac(wk,asm.bad));

		linect = 1;									// Page/SWAP Data Set Counter
		getPageDS(&linect);							// Format Page Data Sets Info
		getSwapDS(&linect);							// Format Swap Data Set Info

		fssRefresh();								// Update Display
		fssSetCursor("COMMAND");					// Default Cursor Position
		fssSetField("MSG1",DASHES);					// Clear Error Msg
		fssSetColor("MSG1",0);						// Default Color

		if(fssGetAID() == fssPFK03 || fssGetAID() == fssPFK15)  // END
			break;
		if(fssGetAID() == fssPFK04 || fssGetAID() == fssPFK16)  // EXIT
			break;
	}

	fssTerm();										// Terminate Full Screen Environemnt

	return 0;
}


//
// Get Aux Storage Manager Counts
//
static void getASMCounts(struct sAsmCounts *asm)
{
	unsigned char *CVT;
	unsigned char *ASMVT;

	CVT   = *( (unsigned char **) (0x10));			// CVT Address
	ASMVT = *( (unsigned char **) (CVT+0x2C0));		// ASMVT Address

	asm->avail = *( (int *) (ASMVT+0x6C));			// Available Slots
	asm->total = *( (int *) (ASMVT+0x70));			// Total Slots
	asm->vio   = *( (int *) (ASMVT+0x74));			// VIO Slots
	asm->adspc = *( (int *) (ASMVT+0x78));			// Address Space Slots
	asm->bad   = *( (int *) (ASMVT+0x7C));			// Bad Slots
}


//
// Get Swap Data Set Info
//
static void getSwapDS(int *i)
{
	unsigned char *CVT;
	unsigned char *ASMVT;
	unsigned char *SART;
	unsigned char *SE;
	unsigned char *DSN;
	unsigned char *UCB;

	char           wk[16];
	char           sd[8];

	char           dsname[44];
	char           volser[6];
	int            size;
	int            avail;
	int            used;
	int            err;

	int            dscnt;
	int            sets;
	int            pages;

	CVT   = *( (unsigned char **) (0x10));			// CVT Address
	ASMVT = *( (unsigned char **) (CVT+0x2C0));		// ASMVT Address
	SART  = *( (unsigned char **) (ASMVT+0x04));	// SART Address
	DSN   = *( (unsigned char **) (SART+0x18));		// Swap Data Set Name 
	SE    = SART+0x50;

	dscnt = *( (int *) (SART+0x08));				// Number of Swap Data Sets
	sets  = *( (int *) (SART+0x24));				// Sets
	pages = *( (int *) (SART+0x38));				// Pages/Set

	//printf("Swap DS=%d Sets=%d Pages/Set=%d\n",dscnt,sets,pages);

	if(dscnt > 0)
	{
		do
		{
			if( (*(SE+9) & 0x80) != 0x80)
			{
				memcpy(dsname,DSN,44);				// Swap Data Set Name
				UCB = *( (unsigned char **) (SE+0x2C));
				memcpy(volser, (UCB+0x1C), 6);		// VOLSER

				size  = *( (int *) (SE+0x10));		// Size
				avail = *( (int *) (SE+0x14));		// Available
				err   = *( (int *) (SE+0x18));		// Error

				used = 100 - ((avail * 100) / size);  // Calc % Used
				if(used == 100)
					used = 99;

				sprintf(wk,"TY%2.2d",*i);			// Format On Current Data Set Line
				fssSetField(wk,"SWAP");

				sprintf(wk,"E%2.2d",*i);
				sprintf(sd,"%3.3d",err);
				fssSetField(wk,sd);

				sprintf(wk,"SZ%2.2d",*i);
				sprintf(sd,"%4.4d",size);
				fssSetField(wk,sd);

				sprintf(wk,"AV%2.2d",*i);
				sprintf(sd,"%4.4d",avail);
				fssSetField(wk,sd);

				sprintf(wk,"U%2.2d",*i);
				sprintf(sd,"%2.2d%%",used);
				fssSetField(wk,sd);

				sprintf(wk,"B%2.2d",*i);
				fssSetField(wk,"");

				sprintf(wk,"VOLS%2.2d",*i);
				fssSetField(wk,volser);

				sprintf(wk,"DSN%2.2d",*i);
				fssSetField(wk,dsname);

				*i = (*i) + 1;
			}

			DSN += 44;								// Next DSN Entry
			SE  += 48;								// Next Part Entry
		} while( *SE != 0x00);
	}
}


//
// Get Page Data Set Info
//
static void getPageDS(int *i)
{
	unsigned char *CVT;
	unsigned char *ASMVT;
	unsigned char *PART;
	unsigned char *DSN;
	unsigned char *PE;
	unsigned char *UCB;

	char           wk[16];
	char           sd[8];

	char           dsname[44];
	char           volser[6];
	char           type[5];
	int            err;
	int            size;
	int            avail;
	int            brst;
	int            used;

	CVT   = *( (unsigned char **) (0x10));			// CVT Address
	ASMVT = *( (unsigned char **) (CVT+0x2C0));		// ASMVT Address
	PART  = *( (unsigned char **) (ASMVT+0x08));	// PART Address
	DSN   = *( (unsigned char **) (PART+0x18));		// Page Data Set Name
	PE    = PART+0x50;								// Page DS Entry

	while( *(PE+0x08) != 0x00 )						// Loop Through until End
	{
		if( (*(PE+9) & 0x80) != 0x80)				// Check for In Use
		{
			memcpy(dsname,DSN,44);					// DSN
			UCB = *( (unsigned char **) (PE+0x2C));
			memcpy(volser, (UCB+0x1C), 6);			// VOLSER
			err       = *( (int *) (PE+0x18));		// Error
			size      = *( (int *) (PE+0x10));		// Size
			avail     = *( (int *) (PE+0x14));		// Available	
			brst      = *( (int *) (PE+0x38));		// Burst Transfer Rate
			brst     /= 256;						//    Divided by 256

			used = 100 - ((avail * 100) / size);	// Calc % Used
			if(used == 100)
				used = 99;

			if( (*(PE+8) & 0x80) == 0x80 )			// Type  of Page Data Set
				strcpy(type,"PLPA");
			else if( (*(PE+8) &0x40) == 0x40)
				strcpy(type,"COMN");
			else if( (*(PE+8) &0x20) == 0x20)
				strcpy(type,"DPLX");
			else
				strcpy(type,"LOCL");

			sprintf(wk,"TY%2.2d",*i);				// Format Information to Screen
			fssSetField(wk,type);

			sprintf(wk,"E%2.2d",*i);
			sprintf(sd,"%3.3d",err);
			fssSetField(wk,sd);

			sprintf(wk,"SZ%2.2d",*i);
			sprintf(sd,"%4.4d",size);
			fssSetField(wk,sd);

			sprintf(wk,"AV%2.2d",*i);
			sprintf(sd,"%4.4d",avail);
			fssSetField(wk,sd);

			sprintf(wk,"U%2.2d",*i);
			sprintf(sd,"%2.2d%%",used);
			fssSetField(wk,sd);

			sprintf(wk,"B%2.2d",*i);
			sprintf(sd,"%3.3d",brst);
			fssSetField(wk,sd);

			sprintf(wk,"VOLS%2.2d",*i);
			fssSetField(wk,volser);

			sprintf(wk,"DSN%2.2d",*i);
			fssSetField(wk,dsname);

			*i = (*i) + 1;
		}
		DSN += 44;    // Next DSN Entry
		PE  += 64;    // Next Part Entry

	}

}


// 
// Get Swap Counts
//
static void getSwapCounts(struct sSwapCounts *sc)
{
	unsigned char *CVT;
	unsigned char *RMCT;
	unsigned char *RMCA;
	unsigned char *WMST;
	int           *Count;


	CVT     = *( (unsigned char **) (0x10));		//  CVT Address
	RMCT    = *( (unsigned char **) (CVT+0x25C));	// RMCT Address
	RMCA    = *( (unsigned char **) (RMCT+0x14));	// RMCA Address
	WMST    = *( (unsigned char **) (RMCT+0x18));	// WMST Address

	memcpy(sc->ips, WMST+4, 2);						// Current IPS Setting
	sc->ips[2] = 0x00;

	Count = (int *) (RMCA+0x7C);					// Address of Counters

	sc->term_wait_in       = *Count;				// Terminal Input
	sc->term_wait_out      = *(Count+1);			// Terminal Output
	sc->long_wait          = *(Count+2);			// Long Wait
	sc->aux_stor_shortage  = *(Count+3);			// Aux Storage
	sc->real_stor_shortage = *(Count+4);			// Real Storage
	sc->detected_wait      = *(Count+5);			// Detected Wait
	sc->request_swap       = *(Count+6);			// Request Swap
	sc->enq_exchange       = *(Count+7);			// Enq Exchange Swap
	sc->workload_exchange  = *(Count+8);			// Workload Swap
	sc->unilateral_swap    = *(Count+9);			// Unilateral Swap
}


//
// Format A Number with Comma Seperators
//
static char *itoacwk;

static void itoac2(int n)
{
	char wk[8];										// Format Work Area


	if(n < 1000)									// If Less Than 1,000
	{
		sprintf(wk,"%d",n);							// Format the value
		strcat(itoacwk,wk);							// Copy to global work
		return;
	}
	itoac2(n/1000);									// Divide by 1,000 & Call Subroutine
	sprintf(wk,",%03d",(n%1000));					// Print Comma & next three digits
	strcat(itoacwk,wk);								// Copy into global work
}


char * itoac(char *buf, int n)
{
	char wk[16];									// Format Work Area

	 itoacwk = wk;									// Work Area Index Pointer
	*itoacwk = 0x00;								// Null String

	if(n < 0)										// Check for Neg Number
	{
		strcat(itoacwk,"-");						// Print Minus Sign
		n = -n;										// Absolute value
	}
	itoac2(n);										// Format Subroutine

	sprintf(buf,"%9.9s",itoacwk);					// Justify

	return buf;
}
