/*-----------------------------------------------------------------------------
|  Copyright (c) 2012-2013, Tommy Sprinkle (tommy@tommysprinkle.com)  
|
|  All rights reserved. Redistribution and use in source and binary forms, 
|  with or without modification, are permitted provided that the following 
|  conditions are met:
|
|    * Redistributions of source code must retain the above copyright notice, 
|      this list of conditions and the following disclaimer. 
|    * Redistributions in binary form must reproduce the above copyright  
|      notice, this list of conditions and the following disclaimer in the  
|      documentation and/or other materials provided with the distribution. 
|    * Neither the name of the author nor the names of its contributors may  
|      be used toendorse or promote products derived from this software  
|      without specific prior written permission. 
|
|   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS  
|   "AS IS" AND ANYEXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
|   LIMITED TO, THE IMPLIED WARRANTIESOF MERCHANTABILITY AND FITNESS  
|   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENTSHALL THE 
|   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
|   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
|   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
|   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
|   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
|   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
|   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
|   OF THE POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "chainr.h"
#include "fss.h"

void fmtUCB(unsigned char * UCB, CHAINR ucbchain);
static int  scanUCB(CHAINR ucbchain);
static struct sUCB * scroll(CHAINR chnr, struct sUCB *ptr, int dir, int amt);
static struct sUCB * locate(CHAINR chnr, struct sUCB *ptr, char *tgt);
static int cmdShift(char * cmd, char *buf);
static int procAID(void);
static int isNumber(char *str);

struct sUCB                         // This record describes one device (UCB)
{
	struct sUCB *next;              // Next record on chain
	struct sUCB *prev;              // Previous record on chain
	int          bcua;              // binary Device Address (CUA)
	char         cua[4];            // characer Device Address
	char         addr[5];           // Address of UCB
	char         type[9];           // Device Type
	char         status[2];         // Device Status
	char         volstat[9];        // Volume Mount Status
	char         volser[7];         // Volume Serial
	char         nrdy[5];           // Device Not Ready Indicator
	char         ainfo[25];         // Additional Device Info
};

#define debug(c) printf("%s\n",c)

#define DASHES "---------------------------------"


int main(int argc, char ** argv)
{
	CHAINR       ucbchain;                          // Anchor for Device Chain
	struct sUCB *scrollPtr;                         // Current Scroll Position
	struct sUCB *urec;                              // Device Record Pointer
	int          i;
	int          sc;
	int          exit_rc;                           // Exit Return Code
	char         fld[16];                           // Field Name Work Area
	char         cmd[80];                           // Command Work Area
	char         wk[80];                            // Command Work Area

	fssInit();                                      // Initialize Full Screen Environment
#include "sucbpnl.pnc"

	fssSetField("MSG1",DASHES);                     // Clear Error Msg
	fssSetColor("MSG1",0);                          // Default Color
	fssSetCursor("COMMAND");                        // Place Cursor in Command Field

	ucbchain = chainNewInt(sizeof(struct sUCB));    // Create Chain

	scanUCB(ucbchain);                              // Scan for UCBs

	scrollPtr = (struct sUCB *) chainFirst(ucbchain);  // Start at top

	do                                              // Terminal Processing Loop
	{
		urec = scrollPtr;                           // Current Scroll Position
		for(i=1; i<22; i++)                         // Max Lines to Display = 22
		{
			if(urec == NULL)                        // All records processed?
				break;

			switch(*(urec->status))                 // Set color based on status
			{
			case 'A':                               // Allocated = White
				sc = fssWHITE;
				break;
			case 'O':                               // Online = Green
				sc = fssGREEN;
				break;
			case 'S':                               // System Console or SYSRES = Yellow
				sc = fssYELLOW;
				break;
			default:                                // Everything Else = Blue
				sc = fssBLUE;
			}

			sprintf(fld,"U%2.2d",i);                // Device Address (CUA)
			fssSetField(fld,urec->cua);
			fssSetColor(fld,sc);

			sprintf(fld,"AD%2.2d",i);               // UCB Address in Storage
			fssSetField(fld,urec->addr);
			fssSetColor(fld,sc);

			sprintf(fld,"TYPE%2.2d",i);             // Device Type
			fssSetField(fld,urec->type);
			fssSetColor(fld,sc);

			sprintf(fld,"STAT%2.2d",i);             // Device Status
			fssSetField(fld,urec->status);
			fssSetColor(fld,sc);

			sprintf(fld,"NR%2.2d",i);               // Not Ready -- Always RED
			fssSetField(fld,urec->nrdy);
			fssSetColor(fld,fssRED);

			sprintf(fld,"VOLS%2.2d",i);             // VOLSER
			fssSetField(fld,urec->volser);
			fssSetColor(fld,sc);

			sprintf(fld,"VOLST%2.2d",i);            // Volume Mount Status
			fssSetField(fld,urec->volstat);
			fssSetColor(fld,sc);

			sprintf(fld,"AINFO%2.2d",i);            // Additional Info
			fssSetField(fld,urec->ainfo);
			fssSetColor(fld,sc);

			urec = urec->next;                      // Next Device Record
		}

		while( i < 22 )                             // If we did not fill up the
		{                                           // page - the pad out rest of
			sprintf(fld,"U%2.2d",i);                // the fields with blanks.
			fssSetField(fld,"");
			sprintf(fld,"AD%2.2d",i);
			fssSetField(fld,"");
			sprintf(fld,"TYPE%2.2d",i);
			fssSetField(fld,"");
			sprintf(fld,"STAT%2.2d",i);
			fssSetField(fld,"");
			sprintf(fld,"NR%2.2d",i);
			fssSetField(fld,"");
			sprintf(fld,"VOLS%2.2d",i);
			fssSetField(fld,"");
			sprintf(fld,"VOLST%2.2d",i);
			fssSetField(fld,"");
			sprintf(fld,"AINFO%2.2d",i);
			fssSetField(fld,"");
			i++;
		}

		fssRefresh();                               // Display Screen, Wait for Input

		fssSetField("MSG1",DASHES);                 // Reset Err Msg 
	    fssSetColor("MSG1",0);                      // Default Color
		fssSetCursor("COMMAND");                    // Default Cursor Position
		procAID();                                  // Process AID
		strcpy(cmd, fssGetField("COMMAND"));        // Copy Command
		if(cmdShift(cmd, wk))                       // Shift off first token
		{
			if(!strcmp("END",wk))                   // END 
			{
				if(cmdShift(cmd,wk))                // No operands allowed
				{
					fssSetField("MSG1", "Invalid Operand for END");
					fssSetColor("MSG1",fssYELLOW);
					fssSetCursor("COMMAND");
				}
				else
				{
					fssReset();                     // Exit...
					exit_rc = 4;
					break;
				}

			}
			if(!strcmp("EXIT",wk))                  // EXIT
			{
				fssReset();                         // Exit...
				exit_rc = 8;
				break;
			}
			else if(!strcmp("DOWN",wk))             // DOWN
			{
				if(cmdShift(cmd,wk))                // Check for Scroll Amt
				{
					strupr(wk);
					if(!strcmp(wk,"MAX") || !strcmp(wk,"M"))   // MAX
					{
						scrollPtr = scroll(ucbchain, scrollPtr, 0, -1);  // Scroll Down Max
						fssSetField("COMMAND","");
					}
					else if(!isNumber(wk))                     // Number Specified ?
					{
						fssSetField("MSG1", "Invalid Scroll Amount");  // No - Invalid
						fssSetColor("MSG1",fssYELLOW);         // Set Error Msg
						fssSetCursor("COMMAND");
					}
					else
					{
						scrollPtr = scroll(ucbchain, scrollPtr, 0, atoi(wk));  // Scroll Down
						fssSetField("COMMAND","");
					}
				}
				else
				{
					scrollPtr = scroll(ucbchain, scrollPtr, 0, 0);   // Default, Scroll 1 Page
					fssSetField("COMMAND","");
				}
			}
			else if(!strcmp("UP",wk))               // UP
			{
				strupr(wk);
				if(cmdShift(cmd,wk))                // Check for Scroll Amt
				{
					if(!strcmp(wk,"MAX") || !strcmp(wk,"M"))  // Max
					{
						scrollPtr = scroll(ucbchain, scrollPtr, 1, -1);  // Scroll Up Max
						fssSetField("COMMAND","");
					}
					else if(!isNumber(wk))                    // If not Number, then Error
					{
						fssSetField("MSG1", "Invalid Scroll Amount");
						fssSetColor("MSG1",fssYELLOW);
						fssSetCursor("COMMAND");
					}
					else
					{
						scrollPtr = scroll(ucbchain, scrollPtr, 1, atoi(wk)); // Scroll Down
						fssSetField("COMMAND","");
					}
				}
				else
				{
					scrollPtr = scroll(ucbchain, scrollPtr, 1, 0);  // Default - One Page
					fssSetField("COMMAND","");
				}
			}
			else if(!strcmp("TOP",wk))              // TOP 
			{
				scrollPtr = scroll(ucbchain, scrollPtr, 1, -1); // Scroll Up Max
				fssSetField("COMMAND","");
			}
			else if(!strcmp("BOT",wk) || !strcmp("BOTTOM",wk))  // BOTTOM
			{
				scrollPtr = scroll(ucbchain, scrollPtr, 0, -1); // Scroll Down Max
				fssSetField("COMMAND","");
			}	
			else if(!strcmp("L",wk))                // LOCATE
			{
				if(!cmdShift(cmd,wk))               // Get operand (device address - cua)
				{
					fssSetField("MSG1","Missing Device Address");
					fssSetColor("MSG1",fssYELLOW);
				}
				scrollPtr = locate(ucbchain, scrollPtr, wk);  // Locate & set Scroll Pos
				fssSetField("COMMAND","");
			}
			else                                             // Else - Invalid Command
			{
				fssSetField("MSG1","Invalid Command");
				fssSetColor("MSG1",fssYELLOW);
				fssSetCursor("COMMAND");
			}
		}
	}
	while(1);

	chainFree(ucbchain);                            // Free Chain Memory

	fssTerm();                                      // Terminate Full Screen Environment

	return exit_rc;                                 // Return To Caller
}

static struct sUCB * locate(CHAINR chnr, struct sUCB *ptr, char *tgt)
{
	struct sUCB *tptr;
	int          itgt;
	int          i;
	char        *p;

	if(!fssIsHex(tgt) || strlen(tgt) != 3)          // CUA - must be 3 hex digits
	{
		fssSetField("MSG1","Invalid Device Address");
		fssSetColor("MSG1",fssYELLOW);
		return ptr;
	}

	strupr(tgt);                                    // Convert To HEX
	p    = tgt;
	itgt = 0;

	for(i=0; i<3; i++)
	{
		itgt = (itgt << 4);
		if(isdigit(*p))
			itgt += *p - '0';
		else
			itgt += (*p - 'A') + 10;
		p++;
	}

	tptr = (struct sUCB *) chainFirst(chnr);        // Start at top of Chain
	while(tptr->next != NULL && tptr->bcua < itgt)  // Scan through looking for a match
		tptr = tptr->next;

	return tptr;
}


static struct sUCB * scroll(CHAINR chnr, struct sUCB *ptr, int dir, int amt)  
{                           //  DIR 0=down, 1=up -- amt 0=page, -1 = max
	int i;

	if(amt == -1)                                   // Scroll MAX
	{
		if(dir > 0)                                 // Up MAX
			return (struct sUCB *) chainFirst(chnr);
		else                                        // Down MAX
		{
			ptr = (struct sUCB *) chainLast(chnr);  // Last entry
			amt = 20;                               // And scroll up 20
			dir = 1;
		}
	}

	if(amt < 1 ||amt > 21)                          // Max Lines is 21
		amt = 21;

	if( dir < 1 && amt == 21 && ptr->next == NULL)  // If at end...
	{
		dir = 1;                                    // Scroll Up 20
		amt = 20;
	}

	if( dir > 0 )                                   // Scroll Up
	{
		for(i=0; i<amt; i++)
		{
			if(ptr->prev != NULL)
				ptr = ptr->prev;
			else
				break;
		}
	}
	else                                            // Scroll Down
	{
		for(i=0; i<amt; i++)
		{
			if(ptr->next != NULL)
				ptr = ptr->next;
			else
				break;
		};
	}

	return ptr;
}


//*****************************
// Add A Record for Each UCB 
//*****************************
static int scanUCB(CHAINR ucbchain)
{
	unsigned char **CVTPTR;
	unsigned char *CVT;

	unsigned char *UCBHW;

	unsigned char *UCB;

	CVTPTR = (unsigned char **) 0x10;               // CVT Pointer x'10'

	CVT = *CVTPTR;                                  // CVT Address

	UCBHW = *((unsigned char **) (CVT+0x28));       // UCB Half Word Table

	while(1)                                        // Loop through Table
	{
		UCB = (unsigned char *) (*UCBHW << 8) + *(UCBHW+1);  // UCB Address
		if( (unsigned int) UCB == 0xFFFF )          // Check for end of UCB List
			break;

		if(UCB != NULL)                             // Ignore if x'0000'
		{
			// format UCB
			fmtUCB(UCB, ucbchain);                  // Build Record for UCB
		}
		UCBHW = UCBHW + 2;                          // Next Entry
	}

	return 0;
}



void fmtUCB(unsigned char * UCB, CHAINR ucbchain)
{
	struct sUCBTYPE                                 // Device Type Record
	{
		unsigned char code;
		char         *type;
	};

	static struct sUCBTYPE URTYPE[] =               // Unit Record Devices
	{
		{ 0x01, "2540R"  },
		{ 0x02, "2540P"  },
		{ 0x03, "2596"   },
		{ 0x04, "2501"   },
		{ 0x05, "2520"   },
		{ 0x06, "3505"   },
		{ 0x08, "1403"   },
		{ 0x09, "3211"   },
		{ 0x0A, "1443"   },
		{ 0x0C, "3525"   },
		{ 0x0E, "3800"   },
		{ 0x10, "2671"   },
		{ 0x16, "3890"   },
		{ 0x18, "2495"   },
		{ 0x1A, "1285"   },
		{ 0x1B, "1287"   },
		{ 0x1C, "1288"   },
		{ 0x1D, "1419"   },
		{ 0x1E, "1419"   },
		{ 0x1F, "1275"   },
		{ 0x20, "1052"   },
		{ 0x22, "3210"   },
		{ 0x23, "3215"   },
		{ 0x30, "3213"   },
		{ 0x42, "3850"   },
		{ 0xFF, "****"   }
	};

	static struct sUCBTYPE TATYPE[] =               // Tape Devices
	{
		{ 0x01, "2400"  },
		{ 0x03, "3420"  },
		{ 0xFF, "****"   }
	};

	static struct sUCBTYPE DATYPE[] =               // DASD Devices
	{
		{ 0x01, "3211"  },
		{ 0x02, "2301"  },
		{ 0x03, "2303"   },
		{ 0x04, "2302"   },
		{ 0x05, "2321"   },
		{ 0x06, "2305"   },
		{ 0x07, "2305"   },
		{ 0x08, "2314"   },
		{ 0x09, "3330-1" },
		{ 0x0A, "3340"   },
		{ 0x0B, "3350"   },
		{ 0x0D, "3330-11"},
		{ 0xFF, "****"   }
	};

	static struct sUCBTYPE GXTYPE[] =               // Graphics Devices
	{
		{ 0x02, "2250"  },
		{ 0x03, "2260"  },
		{ 0x04, "1053"  },
		{ 0x05, "2280"  },
		{ 0x06, "2282"  },
		{ 0x07, "5450"  },
		{ 0x08, "3066"  },
		{ 0x09, "3277"  },
		{ 0x0A, "3284"  },
		{ 0x0B, "3286"  },
		{ 0x0C, "3158"  },
		{ 0xFF, "****"   }
	};

	static struct sUCBTYPE TPTYPE[] =               // TP Devices
	{
		{ 0x93, "2703"  },
		{ 0xFF, "****"   }
	};

	static struct sUCBTYPE CTCTYPE[] =              // CTC Devices
	{
		{ 0xFF, "CTC"   }
	};


	int            i;
	int            bcua;
	char          *type;
	struct sUCB   *ucbRec;
	unsigned char  ut1;
	unsigned char  ut2;

	char UNKWN[8] = "????";
	char status[2];
	char volstat[9];
	char volser[7];
	char nrdy[5];
	char addr[5];
	char cua[4];
	char ainfo[25];

	type = UNKWN;                                   // Don't know device type yet

	ut1 = *(UCB+0x12);                              // Get UCB Device Type Flags
	ut2 = *(UCB+0x13);

	if( ut1 == 0x08)                                // Unit Record Device ?
	{
		for(i=0;;i++)
		{
			if(URTYPE[i].code == 0xFF)              // End of table, not found
			{
				type = URTYPE[i].type;
				break;
			}
			if(URTYPE[i].code == ut2)               // Match
			{
				type = URTYPE[i].type;
				break;
			}
		}
	}
	else if( ut1 == 0x80)                           // Tape Device
	{
		for(i=0;;i++)
		{
			if(TATYPE[i].code == 0xFF)              // End of table, not found
			{
				type = TATYPE[i].type;
				break;
			}
			if(TATYPE[i].code == ut2)               // Match
			{
				type = TATYPE[i].type;
				break;
			}
		}
	}
	else if( ut1 == 0x20)                           // DASD Device
	{
		for(i=0;;i++)
		{
			if(DATYPE[i].code == 0xFF)              // End of table, no match
			{
				type = DATYPE[i].type;
				break;
			}
			if(DATYPE[i].code == ut2)               // Match
			{
				type = DATYPE[i].type;
				break;
			}
		}
	}
	else if( ut1 == 0x10)                           // Graphics Device
	{
		for(i=0;;i++)
		{
			if(GXTYPE[i].code == 0xFF)              // End of table, not found  
			{
				type = GXTYPE[i].type;
				break;
			}
			if(GXTYPE[i].code == ut2)               // Match
			{
				type = GXTYPE[i].type;
				break;
			}
		}
	}
	else if( ut1 == 0x40)                           // TP Device
	{
		for(i=0;;i++)
		{
			if(TPTYPE[i].code == 0xFF)              // End of table, not found
			{
				type = TPTYPE[i].type;
				break;
			}
			if(TPTYPE[i].code == ut2)               // Match
			{
				type = TPTYPE[i].type;
				break;
			}
		}
	}
	else if( ut1 == 0x41)                           // CTC Device
	{
		for(i=0;;i++)
		{
			if(CTCTYPE[i].code == 0xFF)             // End of table, not found
			{
				type = CTCTYPE[i].type;
				break;
			}
			if(CTCTYPE[i].code == ut2)              // Match
			{
				type = CTCTYPE[i].type;
				break;
			}
		}
	}

	strcpy(status,"-");                             // Status not know yet
	memset(volser,0x00,7);                          // Clear VOLSER
	strcpy(volstat,"");                             // Clear Volume Mount Status
	strcpy(ainfo,"");                               // Clear Additional Info

	if( (*(UCB+0x3) & 0x80) == 0x80 )               // Check Device Online
	{
		strcpy(status,"O");                         // Set Status

		if( (*(UCB+0x03) &0x08) == 0x08 )           // Check Device Allocated
			strcpy(status,"A");
		if( (*(UCB+0x03) & 0x02) == 0x02 )          // Check for SYSRES or Console
			strcpy(status,"S");

		if( ut1 == 0x20)                            // If DASD
		{
			if( (*(UCB+0x22) & 0x10) == 0x10 )      // Set Volume Mount Status
				strcpy(volstat,"Private");
			else if( (*(UCB+0x22) & 0x08) == 0x08 )
				strcpy(volstat,"Public");
			else if( (*(UCB+0x22) & 0x04) == 0x04 )
				strcpy(volstat,"Storage");

			// Set Additional Info for DASD (Open DCBs, # Users, # Reserves)
			sprintf(ainfo,"DCB=%3.3d U=%3.3d Rsv=%3.3d",((*(UCB+0x23)) & 0x7F), *(UCB+0x26), *(UCB+0x24)); 
		}

		if( ut1 == 0x20 || ut1 == 0x80)             // DASD or TAPE
		{
			if(*(UCB+0x1C) != 0x00)                 // Check for VOLSER
				memcpy(volser,UCB+0x1c,6);          //   Save it
		}
	}

	// Check for Online & Not Ready
	if( ((*(UCB+0x03) & 0x80) == 0x80) && ((*(UCB+0x06) & 0x40) == 0x40) )
		strcpy(nrdy,"NRDY");
	else
		strcpy(nrdy,"");

	sprintf(addr,"%4.4X",UCB);                      // Format UCB Address
	sprintf(cua,"%3.3s",UCB+0x0D);                  // Format Device Address (CUA)
	bcua = (*(UCB+4) << 8) + *(UCB+5);              // Get Binary Device Address

	ucbRec = (struct sUCB *) chAdd( ucbchain, bcua);  // New Device Record

	strcpy(ucbRec->cua     , cua);                  // -----------------------------
	strcpy(ucbRec->addr    , addr);                 // Copy
	strcpy(ucbRec->type    , type);                 //     In
	strcpy(ucbRec->status  , status);               //       Device 
	strcpy(ucbRec->volstat , volstat);              //             Information
	strcpy(ucbRec->volser  , volser);               //
	strcpy(ucbRec->nrdy    , nrdy);                 // 
	strcpy(ucbRec->ainfo   , ainfo);                // -----------------------------
}


//
//-- shift off a command token
//
static int cmdShift(char * cmd, char *buf)
{
	int   len;
	char *p;
	char *p1;

	p = cmd;                                      // start of command string
	while(*p == ' ')                              // skip over blanks
		p++;

	if(!*p)                                       // if end of string
		return 0;

	p1 = p;                                       // remember start of token
	while(*p && *p != ' ')                        // find end of token
		p++;

	len = p - p1;                                 // token length
	memcpy(buf, p1, len);                         // copy token
	*(buf+len) = '\0';

	strcpy(cmd, p);                               // shift over past removed token
	return 1;
}

//
//-- process AID
//
static int procAID(void)
{
	int    aid;
	char   wk[80];

	wk[0] = '\0';
	aid = fssGetAID();                            // Get AID value
	if( aid == fssPFK03 || aid == fssPFK15)       // PFK 3 or 15
		strcpy(wk, "END ");
	if( aid == fssPFK04 || aid == fssPFK16)       // PFK 4 or 16
		strcpy(wk, "EXIT ");
	else if(aid == fssPFK08 || aid == fssPFK20)   // PFK 8 or 20
		strcpy(wk, "DOWN ");
	else if(aid == fssPFK07 || aid == fssPFK19)   // PFK 7 or 19
		strcpy(wk, "UP ");

	strcat(wk, fssGetField("COMMAND"));           // Append contents of COMMAND field
	fssSetField("COMMAND", strupr(wk));           // Copy back to COMMAND

	return 0;
}


//
//-- Is string all numeric digits
//
static int isNumber(char *str)
{
	while(*str)
	{
		if(!isdigit(*str))
			return 0;
		str++;
	}
	return 1;
}
